# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
import six
import collections
import re
import os
from .support import Status, Defaults, ResourceChanges
from .result import serializeValue, ChangeAware, Results, ResultsMap
from .util import (
    registerClass,
    lookupClass,
    loadModule,
    validateSchema,
    findSchemaErrors,
    UnfurlError,
    UnfurlTaskError,
    UnfurlAddingResourceError,
    filterEnv,
    toEnum,
    wrapSensitiveValue,
)
from . import merge
from .eval import Ref, mapValue, RefContext
from .runtime import RelationshipInstance
from .yamlloader import yaml

import logging

logger = logging.getLogger("unfurl.task")


class TaskRequest(object):
    """
    Yield this to run a child task. (see :py:meth:`unfurl.configurator.TaskView.createSubTask`)
    """

    def __init__(self, configSpec, resource, reason, persist=False, required=None):
        self.configSpec = configSpec
        self.target = resource
        self.reason = reason
        self.persist = persist
        self.required = required
        self.error = configSpec.name == "#error"


class JobRequest(object):
    """
    Yield this to run a child job.
    """

    def __init__(self, resources, errors):
        self.instances = resources
        self.errors = errors

    def __repr__(self):
        return "JobRequest(%s)" % (self.instances,)


# we want ConfigurationSpec to be standalone and easily serializable
class ConfigurationSpec(object):
    @classmethod
    def getDefaults(self):
        return dict(
            className=None,
            majorVersion=0,
            minorVersion="",
            workflow=Defaults.workflow,
            timeout=None,
            operation_host=None,
            environment=None,
            inputs=None,
            inputSchema=None,
            preConditions=None,
            postConditions=None,
            primary=None,
            dependencies=None,
            outputs=None,
        )

    def __init__(
        self,
        name,
        operation,
        className=None,
        majorVersion=0,
        minorVersion="",
        workflow=Defaults.workflow,
        timeout=None,
        operation_host=None,
        environment=None,
        inputs=None,
        inputSchema=None,
        preConditions=None,
        postConditions=None,
        primary=None,
        dependencies=None,
        outputs=None,
    ):
        assert name and className, "missing required arguments"
        self.name = name
        self.operation = operation
        self.className = className
        self.majorVersion = majorVersion
        self.minorVersion = minorVersion
        self.workflow = workflow
        self.timeout = timeout
        self.operationHost = operation_host
        self.environment = environment
        self.inputs = inputs or {}
        self.inputSchema = inputSchema
        self.outputs = outputs or {}
        self.preConditions = preConditions
        self.postConditions = postConditions
        self.artifact = primary

    def findInvalidateInputs(self, inputs):
        if not self.inputSchema:
            return []
        return findSchemaErrors(serializeValue(inputs), self.inputSchema)

    # XXX same for postConditions
    def findInvalidPreconditions(self, target):
        if not self.preConditions:
            return []
        # XXX this should be like a Dependency object
        expanded = serializeValue(target.attributes)
        return findSchemaErrors(expanded, self.preConditions)

    def create(self):
        klass = lookupClass(self.className)
        if not klass:
            raise UnfurlError("Could not load configurator %s" % self.className)
        else:
            return klass(self)

    def shouldRun(self):
        return Defaults.shouldRun

    def copy(self, **mods):
        args = self.__dict__.copy()
        args.update(mods)
        return ConfigurationSpec(**args)

    def __eq__(self, other):
        if not isinstance(other, ConfigurationSpec):
            return False
        # XXX3 add unit tests
        return (
            self.name == other.name
            and self.operation == other.operation
            and self.className == other.className
            and self.majorVersion == other.majorVersion
            and self.minorVersion == other.minorVersion
            and self.workflow == other.workflow
            and self.timeout == other.timeout
            and self.environment == other.environment
            and self.inputs == other.inputs
            and self.inputSchema == other.inputSchema
            and self.outputs == other.outputs
            and self.preConditions == other.preConditions
            and self.postConditions == other.postConditions
        )


class ConfiguratorResult(object):
    """
    Modified indicates whether the underlying state of configuration,
    was changed i.e. the physically altered the system this configuration represents.

    status reports the Status of the current configuration.
    """

    def __init__(
        self,
        success,
        modified,
        status=None,
        configChanged=None,
        result=None,
        outputs=None,
        exception=None,
    ):
        self.modified = modified
        self.status = toEnum(Status, status)
        self.configChanged = configChanged
        self.result = result
        self.success = success
        self.outputs = outputs
        self.exception = None

    def __str__(self):
        result = "" if self.result is None else str(self.result)[:240] + "..."
        return (
            "changes: "
            + (
                " ".join(
                    filter(
                        None,
                        [
                            self.success and "success",
                            self.modified and "modified",
                            self.status is not None and self.status.name,
                        ],
                    )
                )
                or "none"
            )
            + "\n   "
            + result
        )


class AutoRegisterClass(type):
    def __new__(mcls, name, bases, dct):
        cls = type.__new__(mcls, name, bases, dct)
        if cls.shortName:
            name = cls.shortName
        elif name.endswith("Configurator"):
            name = name[: -len("Configurator")]
        if name:
            registerClass(cls.__module__ + "." + cls.__name__, cls, name)
        return cls


@six.add_metaclass(AutoRegisterClass)
class Configurator(object):

    shortName = None
    """shortName can be used to customize the "short name" of the configurator
    as an alternative to using the full name ("module.class") when setting the implementation on an operation.
    (Titlecase recommended)"""

    def __init__(self, configurationSpec):
        self.configSpec = configurationSpec

    def getGenerator(self, task):
        return self.run(task)

    # yields a JobRequest, TaskRequest or a ConfiguratorResult
    def run(self, task):
        """
        This should perform the operation specified in the :class:`ConfigurationSpec`
        on the :obj:`task.target`.

        Args:
            task (:class:`TaskView`) The task currently running.

        Yields:
            Should yield either a :class:`JobRequest`, :class:`TaskRequest`
            or a :class:`ConfiguratorResult` when done
        """
        yield task.done(False)

    def canDryRun(self, task):
        """
        Returns whether this configurator can handle a dry-runs for the given task.
        (And should check `task.dryRun` in during run()"".

        Args:
            task (:obj:`TaskView`) The task about to be run.

        Returns:
            bool
        """
        return False

    def canRun(self, task):
        """
        Return whether or not the configurator can execute the given task.

        Does this configurator support the requested action and parameters
        and given the current state of the target instance?

        Args:
            task (:class:`TaskView`) The task that is about to be run.

        Returns:
            (bool or str): Should return True or a message describing why the task couldn't be run.
        """
        return True

    def shouldRun(self, task):
        """Does this configuration need to be run?"""
        return self.configSpec.shouldRun()

    # XXX3 should be called during when checking dependencies
    # def checkConfigurationStatus(self, task):
    #   """Is this configuration still valid?"""
    #   return Status.ok


class TaskView(object):
    """The interface presented to configurators."""

    def __init__(self, manifest, configSpec, target, reason=None, dependencies=None):
        # public:
        self.configSpec = configSpec
        self.target = target
        self.reason = reason
        self.logger = logger
        self.cwd = os.path.abspath(self.target.baseDir)
        # private:
        self._errors = []  # UnfurlTaskError objects appends themselves to this list
        self._inputs = None
        self._environ = None
        self._manifest = manifest
        self.messages = []
        self._addedResources = []
        self._dependenciesChanged = False
        self.dependencies = dependencies or {}
        self._resourceChanges = ResourceChanges()
        # public:
        self.operationHost = self._findOperationHost(target, configSpec.operationHost)

    @property
    def inputs(self):
        """
        Exposes inputs and task settings as expression variables, so they can be accessed like:

        eval: $inputs::param

        or in jinja2 templates:

        {{ inputs.param }}
        """
        if self._inputs is None:
            inputs = self.configSpec.inputs.copy()
            relationship = isinstance(self.target, RelationshipInstance)
            if relationship:
                target = self.target.target
            else:
                target = self.target
            HOST = (target.parent or target).attributes
            ORCHESTRATOR = target.root.findInstanceOrExternal("localhost")
            vars = dict(
                inputs=inputs,
                task=self.getSettings(),
                connections=list(self._getConnections()),
                allConnections=self._getAllConnections(),
                SELF=self.target.attributes,
                HOST=HOST,
                ORCHESTRATOR=ORCHESTRATOR and ORCHESTRATOR.attributes or {},
                OPERATION_HOST=self.operationHost
                and self.operationHost.attributes
                or {},
            )
            if relationship:
                vars["SOURCE"] = self.target.source.attributes
                vars["TARGET"] = target.attributes
            # expose inputs lazily to allow self-referencee
            ctx = RefContext(self.target, vars)
            if self.configSpec.artifact and self.configSpec.artifact.baseDir:
                ctx.baseDir = self.configSpec.artifact.baseDir
            self._inputs = ResultsMap(inputs, ctx)
        return self._inputs

    @property
    def vars(self):
        """
        A dictionary of the same variables that are available to expressions when evaluating inputs.
        """
        return self.inputs.context.vars

    def _getConnections(self):
        seen = set()
        for parent in reversed(self.target.ancestors):
            # use reversed() so nearer overrides farther
            # XXX broken if multiple requirements point to same parent (e.g. dev and prod connections)
            # XXX test if operationHost is external (e.g locahost) getRequirements() matches local parent
            found = False
            if self.operationHost:
                for rel in self.operationHost.getRequirements(parent):
                    # examine both the relationship's properties and its capability's properties
                    found = True
                    if id(rel) not in seen:
                        seen.add(id(rel))
                        yield rel

            if not found:
                # not found, see if there's a default connection
                # XXX this should use the same relationship type as findConnection()
                for rel in parent.getDefaultRelationships():
                    if id(rel) not in seen:
                        seen.add(id(rel))
                        yield rel

    def _findRelationshipEnvVars(self):
        """
        We look for instances that the task's implementation might to connect to
        (by following the targets hostedOn relationships)
        and check if the operation_host has any relationships with those instances too.
        If it does, collect any environment variables set by those connections.

        For example, consider an implementation whose target is a Kubernetes cluster hosted on GCP.
        The operation_host's connections to those instances might set KUBECONFIG and GOOGLE_APPLICATION_CREDENTIALS
        respectively and the implementation will probably need both those set when it executes.
        """
        env = {}
        t = lambda datatype: datatype.type == "unfurl.datatypes.EnvVar"
        # XXX broken if multiple requirements point to same parent (e.g. dev and prod connections)
        for rel in self._getConnections():
            env.update(rel.mergeProps(t))

        return env

    def getEnvironment(self, addOnly):
        # note: inputs should be evaluated before environment
        # use merge.copy to preserve basedir
        rules = merge.copy(self.target.root.envRules)

        if self.configSpec.environment:
            rules.update(self.configSpec.environment)
        rules = serializeValue(
            mapValue(rules, self.inputs.context), resolveExternal=True
        )
        env = filterEnv(rules, addOnly=addOnly)
        relEnvVars = self._findRelationshipEnvVars()
        env.update(relEnvVars)
        targets = []
        if isinstance(self.target, RelationshipInstance):
            targets = [
                c.tosca_id
                for c in self.target.target.getCapabilities(
                    self.target.capability.template.name
                )
            ]
            env.update(
                dict(
                    TARGETS=",".join(targets),
                    TARGET=self.target.target.tosca_id,
                    SOURCES=",".join(
                        [
                            r.tosca_id
                            for r in self.target.source.getRequirements(
                                self.target.requirement.template.name
                            )
                        ]
                    ),
                    SOURCE=self.target.source.tosca_id,
                )
            )
        return env

    def getSettings(self):
        return dict(
            verbose=self.verbose,
            name=self.configSpec.name,
            dryrun=self.dryRun,
            workflow=self.configSpec.workflow,
            operation=self.configSpec.operation,
            timeout=self.configSpec.timeout,
            target=self.target.name,
            reason=self.reason,
            cwd=self.cwd,
        )

    def _findOperationHost(self, target, operation_host):
        # SELF, HOST, ORCHESTRATOR, SOURCE, TARGET
        if not operation_host or operation_host in ["localhost", "ORCHESTRATOR"]:
            return target.root.findInstanceOrExternal("localhost")
        if operation_host == "SELF":
            return target
        if operation_host == "HOST":
            # XXX should search all ancestors to find parent that can handle the given operation
            # e.g. ansible configurator should find ancestor compute node
            return target.parent
        if operation_host == "SOURCE":
            return target.source
        if operation_host == "TARGET":
            return target.target
        host = target.root.findInstanceOrExternal(operation_host)
        if host:
            return host
        raise UnfurlTaskError(self, "can not find operation_host: %s" % operation_host)

    def _getAllConnections(self):
        cons = {}
        if self.operationHost:
            for rel in self.operationHost.requirements:
                cons.setdefault(rel.name, []).append(rel)
        for rel in self.target.root.requirements:
            if rel.name not in cons:
                cons[rel.name] = [rel]

        return cons

    def findConnection(self, target, relation="tosca.relationships.ConnectsTo"):
        connection = self.query(
            "$OPERATION_HOST::.requirements::*[.type=%s][.target=$target]" % relation,
            vars=dict(target=target),
        )
        # alternative query: [.type=unfurl.nodes.K8sCluster]::.capabilities::.relationships::[.type=unfurl.relationships.ConnectsTo.K8sCluster][.source=$OPERATION_HOST]
        if not connection:
            # no connection, see if there's a default relationship template defined for this target
            endpoints = target.getDefaultRelationships(relation)
            if endpoints:
                connection = endpoints[0]
        return connection

    def sensitive(self, value):
        """Mark the given value as sensitive. Sensitive values will be encrypted or redacted when outputed.

        Returns:
          sensitive: A subtype of `sensitive` appropriate for the value or the value itself if it can't be converted.

        """
        return wrapSensitiveValue(
            value, self.operationHost and self.operationHost.templar._loader._vault
        )

    def addMessage(self, message):
        self.messages.append(message)

    def findInstance(self, name):
        return self._manifest.getRootResource().findInstanceOrExternal(name)

    # XXX
    # def pending(self, modified=None, sleep=100, waitFor=None, outputs=None):
    #     """
    #     >>> yield task.pending(60)
    #
    #     set modified to True to advise that target has already been modified
    #
    #     outputs to share operation outputs so far
    #     """

    def done(
        self,
        success=None,
        modified=None,
        status=None,
        result=None,
        outputs=None,
        captureException=None,
    ):
        """`run()` should call this method and yield its return value before terminating.

        >>> yield task.done(True)

        Args:

          success (bool):  indicates if this operation completed without an error.
          modified (bool): (optional) indicates whether the physical instance was modified by this operation.
          status (Status): (optional) should be set if the operation changed the operational status of the target instance.
                   If not specified, the runtime will updated the instance status as needed, based
                   the operation preformed and observed changes to the instance (attributes changed).
          result (dict):  (optional) A dictionary that will be serialized as YAML into the changelog, can contain any useful data about these operation.
          outputs (dict): (optional) Operation outputs, as specified in the toplogy template.

        Returns:
              :class:`ConfiguratorResult`
        """
        if success is None:
            success = not self._errors
        if isinstance(modified, Status):
            status = modified
            modified = True

        kw = dict(result=result, outputs=outputs)
        if captureException is not None:
            logLevel = logging.DEBUG if success else logging.ERROR
            kw["exception"] = UnfurlTaskError(self, captureException, logLevel)

        return ConfiguratorResult(success, modified, status, **kw)

    # updates can be marked as dependencies (changes to dependencies changed) or required (error if changed)
    # configuration has cumulative set of changes made it to resources
    # updates update those changes
    # other configurations maybe modify those changes, triggering a configuration change
    def query(
        self,
        query,
        dependency=False,
        name=None,
        required=False,
        wantList=False,
        resolveExternal=True,
        strict=True,
        vars=None,
    ):
        # XXX pass resolveExternal to context?
        try:
            result = Ref(query, vars=vars).resolve(
                self.inputs.context, wantList, strict
            )
        except:
            UnfurlTaskError(
                self, "error while evaluating query: %s" % query, logging.WARNING
            )
            return None

        if dependency:
            self.addDependency(
                query, result, name=name, required=required, wantList=wantList
            )
        return result

    def addDependency(
        self,
        expr,
        expected=None,
        schema=None,
        name=None,
        required=False,
        wantList=False,
    ):
        getter = getattr(expr, "asRef", None)
        if getter:
            # expr is a configuration or resource or ExternalValue
            expr = Ref(getter()).source

        dependency = Dependency(expr, expected, schema, name, required, wantList)
        self.dependencies[name or expr] = dependency
        self.dependenciesChanged = True
        return dependency

    def removeDependency(self, name):
        old = self.dependencies.pop(name, None)
        if old:
            self.dependenciesChanged = True
        return old

    # def createConfigurationSpec(self, name, configSpec):
    #     if isinstance(configSpec, six.string_types):
    #         configSpec = yaml.load(configSpec)
    #     return self._manifest.loadConfigSpec(name, configSpec)

    def createSubTask(
        self, operation, resource=None, inputs=None, persist=False, required=False
    ):
        """Create a subtask that will be executed if yielded by `run()`

        Args:
          operation (str): The operation call (like `interface.operation`)
          resource (:class:`NodeInstance`) The current target if missing.

        Returns:
           :class:`TaskRequest`
        """
        if resource is None:
            resource = self.target

        if inputs is None:
            inputs = self.configSpec.inputs

        if isinstance(operation, six.string_types):
            taskRequest = self.job.plan.createTaskRequest(
                operation, resource, "for subtask: " + self.configSpec.name, inputs
            )
            if taskRequest.error:
                return None
            else:
                taskRequest.persist = persist
                taskRequest.required = required
                return taskRequest

        # XXX:
        # # Configurations created by subtasks are transient insofar as the are not part of the spec,
        # # but they are recorded as part of the resource's configuration state.
        # # Marking as persistent or required will create a dependency on the new configuration.
        # if persist or required:
        #  expr = "::%s::.configurations::%s" % (configSpec.target, configSpec.name)
        #  self.addDependency(expr, required=required)

        # operation should be a ConfigurationSpec
        return TaskRequest(operation, resource, "subtask", persist, required)

    # # XXX how can we explicitly associate relations with target resources etc.?
    # # through capability attributes and dependencies/relationship attributes
    def updateResources(self, resources):
        """Notifies Unfurl of new or changes to instances made while the configurator was running.

        Operational status indicates if the instance currently exists or not.
        This will queue a new child job if needed.

        .. code-block:: YAML

          - name:     aNewResource
            template: aNodeTemplate
            parent:   HOST
            attributes:
               anAttribute: aValue
            readyState:
              local: ok
              state: state
          - name:     SELF
            attributes:
                anAttribute: aNewValue

        Args:
          resources (list or str): Either a list or string that is parsed as YAML.

        Returns:
          :class:`JobRequest`: To run the job based on the supplied spec
              immediately, yield the returned JobRequest.
        """
        from .manifest import Manifest

        if isinstance(resources, six.string_types):
            try:
                resources = yaml.load(resources)
            except:
                UnfurlTaskError(self, "unable to parse as YAML: %s" % resources)
                return None

        if isinstance(resources, collections.Mapping):
            resources = [resources]
        elif not isinstance(resources, collections.MutableSequence):
            UnfurlTaskError(
                self,
                "updateResources requires a list of updates, not a %s"
                % type(resources),
            )
            return None

        errors = []
        newResources = []
        newResourceSpecs = []
        for resourceSpec in resources:
            originalResourceSpec = resourceSpec
            try:
                rname = resourceSpec.get("name", "SELF")
                if rname == ".self" or rname == "SELF":
                    existingResource = self.target
                else:
                    existingResource = self.findInstance(rname)

                if existingResource:
                    # XXX2 if spec is defined (not just status), there should be a way to
                    # indicate this should replace an existing resource or throw an error
                    if "readyState" in resourceSpec:
                        # we need to set this explicitly for the attribute manager to track status
                        # XXX track all status attributes (esp. state and created) and remove this hack
                        operational = Manifest.loadStatus(resourceSpec)
                        if operational.localStatus is not None:
                            existingResource.localStatus = operational.localStatus
                        if operational.state is not None:
                            existingResource.state = operational.state

                    attributes = resourceSpec.get("attributes")
                    if attributes:
                        for key, value in mapValue(
                            attributes, existingResource
                        ).items():
                            existingResource.attributes[key] = value
                            logger.debug(
                                "setting attribute %s with %s on %s",
                                key,
                                value,
                                existingResource.name,
                            )
                    logger.info("updating resources %s", existingResource.name)
                    continue

                pname = resourceSpec.get("parent")
                if pname in [".self", "SELF"]:
                    resourceSpec["parent"] = self.target.name
                elif pname == "HOST":
                    resourceSpec["parent"] = (
                        self.target.parent.name if self.target.parent else "root"
                    )

                if isinstance(resourceSpec.get("template"), dict):
                    # inline node template, add it to the spec
                    tname = resourceSpec["template"].pop("name", rname)
                    nodeSpec = self._manifest.tosca.addNodeTemplate(
                        tname, resourceSpec["template"]
                    )
                    resourceSpec["template"] = nodeSpec.name

                if resourceSpec.get("readyState") and "created" not in resourceSpec:
                    # setting "created" to the target's key indicates that
                    # the target is responsible for deletion
                    # if "created" is not defined, set it if readyState is set
                    resourceSpec["created"] = self.target.key

                if (
                    self.job
                    and "parent" not in resourceSpec
                    and "template" in resourceSpec
                ):
                    nodeSpec = self._manifest.tosca.getTemplate(
                        resourceSpec["template"]
                    )
                    parent = (
                        self.job.plan.findParentResource(nodeSpec) or self.target.root
                    )
                else:
                    parent = self.target.root
                # note: if resourceSpec[parent] is set it overrides the parent keyword
                resource = self._manifest.createNodeInstance(
                    rname, resourceSpec, parent=parent
                )

                # XXX wrong... these need to be operational instances
                # if resource.required or resourceSpec.get("dependent"):
                #    self.addDependency(resource, required=resource.required)
            except:
                errors.append(UnfurlAddingResourceError(self, originalResourceSpec))
            else:
                newResourceSpecs.append(originalResourceSpec)
                newResources.append(resource)

        if newResourceSpecs:
            self._resourceChanges.addResources(newResourceSpecs)
            self._addedResources.extend(newResources)
            logger.info("add resources %s", newResources)

            jobRequest = JobRequest(newResources, errors)
            if self.job:
                self.job.jobRequestQueue.append(jobRequest)
            return jobRequest
        return None


class Dependency(ChangeAware):
    """Represents a runtime dependency for a configuration.


      Dependencies are used to determine if a configuration needs re-run as follows:

    * Tosca `DependsOn`

      * They are dynamically created when evaluating and comparing the
        configuration spec's attributes with the previous values

      * Persistent dependencies can be created when the configurator
        invokes these apis: `createSubTask`, `updateResources`, `query`, `addDependency`
    """

    def __init__(
        self,
        expr,
        expected=None,
        schema=None,
        name=None,
        required=False,
        wantList=False,
    ):
        """
        if schema is not None, validate the result using schema
        if expected is not None, test that result equals expected
        otherwise test that result isn't empty has not changed since the last attempt
        """
        assert not (expected and schema)
        self.expr = expr

        self.expected = expected
        self.schema = schema
        self.required = required
        self.name = name
        self.wantList = wantList

    def refresh(self, config):
        if self.expected is not None:
            changeId = config.changeId
            context = RefContext(
                config.target, dict(val=self.expected, changeId=changeId)
            )
            result = Ref(self.expr).resolve(context, wantList=self.wantList)
            self.expected = result

    @staticmethod
    def hasValueChanged(value, changeset):
        if isinstance(value, Results):
            return Dependency.hasValueChanged(value._attributes, changeset)
        elif isinstance(value, collections.Mapping):
            if any(Dependency.hasValueChanged(v, changeset) for v in value.values()):
                return True
        elif isinstance(value, (collections.MutableSequence, tuple)):
            if any(Dependency.hasValueChanged(v, changeset) for v in value):
                return True
        elif isinstance(value, ChangeAware):
            return value.hasChanged(changeset)
        else:
            return False

    def hasChanged(self, config):
        changeId = config.changeId
        context = RefContext(config.target, dict(val=self.expected, changeId=changeId))
        result = Ref(self.expr).resolveOne(context)  # resolve(context, self.wantList)

        if self.schema:
            # result isn't as expected, something changed
            if not validateSchema(result, self.schema):
                return False
        else:
            if self.expected is not None:
                expected = mapValue(self.expected, context)
                if result != expected:
                    logger.debug("hasChanged: %s != %s", result, expected)
                    return True
            elif not result:
                # if expression no longer true (e.g. a resource wasn't found), then treat dependency as changed
                return True

        if self.hasValueChanged(result, config):
            return True

        return False


def _setDefaultCommand(kw, implementation, inputs):
    # is it a shell script or a command line?
    shell = inputs.get("shell")
    if shell is None:
        # no special shell characters
        shell = not re.match(r"[\w.-]+\Z", implementation)

    operation_host = kw.get("operation_host")
    implementation = implementation.lstrip()
    if not operation_host or operation_host == "localhost":
        className = "unfurl.configurators.shell.ShellConfigurator"
        if shell:
            shellArgs = dict(command=implementation)
        else:
            shellArgs = dict(command=[implementation])
    else:
        className = "unfurl.configurators.ansible.AnsibleConfigurator"
        module = "shell" if shell else "command"
        playbookTask = dict(cmd=implementation)
        cwd = inputs.get("cwd")
        if cwd:
            playbookTask["chdir"] = cwd
        if shell and isinstance(shell, six.string_types):
            playbookTask["executable"] = shell
        shellArgs = dict(playbook=[{module: playbookTask}])

    kw["className"] = className
    if inputs:
        shellArgs.update(inputs)
    kw["inputs"] = shellArgs


def getConfigSpecArgsFromImplementation(iDef, inputs, template):
    # XXX template should be operation_host's template!
    implementation = iDef.implementation
    kw = dict(inputs=inputs, outputs=iDef.outputs)
    configSpecArgs = ConfigurationSpec.getDefaults()
    artifact = None
    if isinstance(implementation, dict):
        for name, value in implementation.items():
            if name == "primary":
                artifact = template.findOrCreateArtifact(value, path=iDef._source)
            elif name == "dependencies":
                kw[name] = [
                    template.findOrCreateArtifact(artifactTpl, path=iDef._source)
                    for artifactTpl in value
                ]
            elif name in configSpecArgs:
                kw[name] = value

    else:
        # "either because it refers to a named artifact specified in the artifacts section of a type or template,
        # or because it represents the name of a script in the CSAR file that contains the definition."
        artifact = template.findOrCreateArtifact(implementation, path=iDef._source)
    kw["primary"] = artifact
    assert artifact or "className" in kw

    if "className" not in kw:
        if not artifact:  # malformed implementation
            return None
        implementation = artifact.file
        try:
            # see if implementation looks like a python class
            if "#" in implementation:
                path, fragment = artifact.getPathAndFragment()
                mod = loadModule(path)
                kw["className"] = mod.__name__ + "." + fragment
                return kw
            elif lookupClass(implementation):
                kw["className"] = implementation
                return kw
        except:
            pass
        # assume it's a command line
        logger.debug(
            "interpreting 'implementation' as a shell command: %s", implementation
        )
        _setDefaultCommand(kw, implementation, inputs)
    return kw



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
import sys

# from unfurl.eval import Ref # broken in 2.7

# from ansible.errors import AnsibleError

# module name is lookup name
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):
        # resource should be current host or current config if no host
        refContext = variables["__unfurl"]
        # workaround for 2.7
        Ref = sys.modules[type(variables["__unfurl"]).__module__].Ref

        return list(map(lambda term: Ref(term).resolveOne(refContext), terms))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError
from toscaparser.elements.statefulentitytype import StatefulEntityType

SECTIONS = (
    LIFECYCLE,
    CONFIGURE,
    INSTALL,
    LIFECYCLE_SHORTNAME,
    CONFIGURE_SHORTNAME,
    INSTALL_SHORTNAME,
) = (
    "tosca.interfaces.node.lifecycle.Standard",
    "tosca.interfaces.relationship.Configure",
    "unfurl.interfaces.Install",
    "Standard",
    "Configure",
    "Install",
)

OPERATION_DEF_RESERVED_WORDS = (DESCRIPTION, IMPLEMENTATION, INPUTS, OUTPUTS) = (
    "description",
    "implementation",
    "inputs",
    "outputs",
)

INTERFACE_DEF_RESERVED_WORDS = ["type", "inputs", "operations", "notifications", "description", "implementation"]

# this is kind of misnamed, these are created for each operation defined on a interface definition
class InterfacesDef(StatefulEntityType):
    """TOSCA built-in interfaces type."""

    def __init__(
        self,
        node_type,
        interfacetype,
        node_template=None,
        name=None,
        value=None,
        inputs=None,
        outputs=None
    ):
        self.ntype = node_type
        self.node_template = node_template
        self.iname = interfacetype
        self.name = name
        self.value = value
        self.implementation = None
        self.inputs = inputs
        self._source = None
        self.outputs = outputs
        self.defs = {}
        interfaces = getattr(self.ntype, "interfaces", None)
        if "type" in interfaces.get(interfacetype, {}):
            interfacetype = interfaces[interfacetype]["type"]
        elif interfacetype == LIFECYCLE_SHORTNAME:
            interfacetype = LIFECYCLE
        elif interfacetype == CONFIGURE_SHORTNAME:
            interfacetype = CONFIGURE
        elif interfacetype == INSTALL_SHORTNAME:
            interfacetype = INSTALL
        self.type = interfacetype
        if node_type:
            if (
                self.node_template
                and self.node_template.custom_def
                and interfacetype in self.node_template.custom_def
            ):
                self.defs = self.node_template.custom_def[interfacetype]
            else:
                self.defs = self.TOSCA_DEF[interfacetype]
        if value:
            if isinstance(self.value, dict):
                for i, j in self.value.items():
                    if i == '_source':
                        self._source = j
                    elif i == IMPLEMENTATION:
                        self.implementation = j
                    elif i == INPUTS:
                        if self.inputs:
                            self.inputs.update(j)
                        else:
                            self.inputs = j
                    elif i == OUTPUTS:
                        if self.outputs:
                            self.outputs.update(j)
                        else:
                            self.outputs = j
                    elif i not in OPERATION_DEF_RESERVED_WORDS:
                        what = '"interfaces" of template "%s"' % self.node_template.name
                        ExceptionCollector.appendException(
                            UnknownFieldError(what=what, field=i)
                        )
            else:
                self.implementation = value

    @property
    def lifecycle_ops(self):
        if self.defs:
            if self.type == LIFECYCLE:
                return self._ops()

    @property
    def configure_ops(self):
        if self.defs:
            if self.type == CONFIGURE:
                return self._ops()

    def _ops(self):
        ops = []
        for name in list(self.defs.keys()):
            ops.append(name)
        return ops



# SPDX-License-Identifier: MIT
# Copyright (c) 2020 Adam Souzis
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import pbr.version


def __version__(release=False):
    # a function because this is expensive
    if release:
        return pbr.version.VersionInfo(__name__).release_string()
    else:
        return pbr.version.VersionInfo(__name__).version_string()


import os
import os.path
import sys

vendor_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "vendor")
sys.path.insert(0, vendor_dir)


class DefaultNames(object):
    SpecDirectory = "spec"
    EnsembleDirectory = "ensemble"
    Ensemble = "ensemble.yaml"
    EnsembleTemplate = "ensemble-template.yaml"
    ServiceTemplate = "service-template.yaml"
    LocalConfig = "unfurl.yaml"
    HomeDirectory = ".unfurl_home"
    JobsLog = "jobs.tsv"


def getHomeConfigPath(homepath):
    # if homepath is explicitly it overrides UNFURL_HOME
    # (set it to empty string to disable the homepath)
    # otherwise use UNFURL_HOME or the default location
    if homepath is None:
        if "UNFURL_HOME" in os.environ:
            homepath = os.getenv("UNFURL_HOME")
        else:
            homepath = os.path.join("~", DefaultNames.HomeDirectory)
    if homepath:
        homepath = os.path.expanduser(homepath)
        if not os.path.exists(homepath):
            isdir = not homepath.endswith(".yml") and not homepath.endswith(".yaml")
        else:
            isdir = os.path.isdir(homepath)
        if isdir:
            return os.path.abspath(os.path.join(homepath, DefaultNames.LocalConfig))
        else:
            return os.path.abspath(homepath)
    return None


### logging initialization
import logging

logging.captureWarnings(True)
_logHandler = None


class sensitive(object):
    redacted_str = "<<REDACTED>>"

    def __sensitive__(self):
        return True


class SensitiveLogFilter(logging.Filter):
    def filter(self, record):
        # redact any sensitive value
        record.args = tuple(
            sensitive.redacted_str if isinstance(a, sensitive) else a
            for a in record.args
        )
        return True


# def addFileLogger():


def initLogging(level=None, logfile=None):
    rootLogger = logging.getLogger()
    oldLevel = rootLogger.getEffectiveLevel()
    f = SensitiveLogFilter()

    # global _logHandler so we can call initLogging multiple times
    global _logHandler
    if not _logHandler:
        rootLogger.setLevel(logging.DEBUG)  # need to set this first
        rootLogger.addFilter(f)

        _logHandler = logging.StreamHandler()
        formatter = logging.Formatter("%(name)s:%(levelname)s: %(message)s")
        _logHandler.setFormatter(formatter)
        _logHandler.addFilter(f)
        rootLogger.addHandler(_logHandler)

    if logfile:
        ch = logging.FileHandler(logfile)
        formatter = logging.Formatter(
            "[%(asctime)s] %(name)s:%(levelname)s: %(message)s"
        )
        ch.setFormatter(formatter)
        ch.setLevel(logging.DEBUG)
        ch.addFilter(f)
        rootLogger.addHandler(ch)

    if level is not None:
        _logHandler.setLevel(level)
    return _logHandler, oldLevel


_logEnv = os.getenv("UNFURL_LOGGING")
if _logEnv is not None:
    initLogging(_logEnv.upper())

### Ansible initialization
if "ANSIBLE_CONFIG" not in os.environ:
    os.environ["ANSIBLE_CONFIG"] = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "configurators", "ansible.cfg")
    )
try:
    import ansible
except ImportError:
    pass
else:
    import ansible.constants as C

    if "ANSIBLE_NOCOWS" not in os.environ:
        C.ANSIBLE_NOCOWS = 1
    if "ANSIBLE_JINJA2_NATIVE" not in os.environ:
        C.DEFAULT_JINJA2_NATIVE = 1

    import ansible.utils.display

    ansible.utils.display.logger = logging.getLogger("unfurl.ansible")
    display = ansible.utils.display.Display()

    # Display is a singleton which we can't subclass so monkey patch instead
    _super_display = ansible.utils.display.Display.display

    def _display(self, msg, color=None, stderr=False, screen_only=False, log_only=True):
        if screen_only:
            return
        return _super_display(self, msg, color, stderr, screen_only, log_only)

    ansible.utils.display.Display.display = _display

    from ansible.plugins.loader import lookup_loader, filter_loader

    lookup_loader.add_directory(os.path.abspath(os.path.dirname(__file__)), True)
    filter_loader.add_directory(os.path.abspath(os.path.dirname(__file__)), True)



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
"""
enviroment:
timeout:
inputs:
 command: "--switch {{ '.::foo' | ref }}"
 cwd
 dryrun
 shell
 keeplines
 done
 resultTemplate: # available vars: cmd, stdout, stderr, returncode, error, timeout
   eval:
    file:
      ./handleResult.tpl
   foreach: contents # get the file contents
"""


# see also 13.4.1 Shell scripts p 360
# XXX add support for a stdin parameter

from ..configurator import Status
from ..util import which
from . import TemplateConfigurator
import os
import sys
import shlex
import six
from click.termui import unstyle

if os.name == "posix" and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess
# cf https://github.com/opsmop/opsmop/blob/master/opsmop/core/command.py


class _PrintOnAppendList(list):
    def append(self, data):
        list.append(self, data)
        try:
            sys.stdout.write(data.decode())
        except:
            pass


def _run(*args, **kwargs):
    timeout = kwargs.pop("timeout", None)
    input = kwargs.pop("input", None)
    with subprocess.Popen(*args, **kwargs) as process:
        try:
            stdout = None
            stderr = None
            _save_input = process._save_input
            # _save_input is called after _fileobj2output is setup but before reading
            def _save_input_hook_hack(input):
                if process.stdout:
                    process._fileobj2output[process.stdout] = _PrintOnAppendList()
                if process.stderr:
                    process._fileobj2output[process.stderr] = _PrintOnAppendList()
                _save_input(input)

            process._save_input = _save_input_hook_hack
            process.communicate(input, timeout=timeout)
            if process.stdout:
                stdout = b"".join(process._fileobj2output[process.stdout])
            if process.stderr:
                stderr = b"".join(process._fileobj2output[process.stderr])
        except subprocess.TimeoutExpired:
            process.kill()
            process.wait()
            raise
        except:  # Including KeyboardInterrupt, communicate handled that.
            process.kill()
            # We don't call process.wait() as .__exit__ does that for us.
            raise
        retcode = process.poll()
    return subprocess.CompletedProcess(process.args, retcode, stdout, stderr)


def _truncate(s):
    if len(s) > 1000:
        return "%s [%s omitted...]  %s" % (s[:500], len(s), s[-500:])
    return s


# XXX we should know if cmd if not os.access(implementation, os.X):
class ShellConfigurator(TemplateConfigurator):
    _defaultCmd = None

    @staticmethod
    def _cmd(cmd, keeplines):
        if not isinstance(cmd, six.string_types):
            cmdStr = " ".join(cmd)
        else:
            if not keeplines:
                cmd = cmd.replace("\n", " ")
            cmdStr = cmd
            cmd = shlex.split(cmd)
        return cmdStr, cmd

    def runProcess(
        self,
        cmd,
        shell=False,
        timeout=None,
        env=None,
        cwd=None,
        keeplines=False,
        echo=True,
    ):
        """
        Returns an object with the following attributes:

        cmd
        timeout (None unless timeout occurred)
        stderr
        stdout
        returncode (None if the process didn't complete)
        error if an exception was raised
        """
        cmdStr, cmd = self._cmd(cmd, keeplines)
        try:
            # hack to echo results
            if echo and hasattr(subprocess.Popen, "_save_input"):
                run = _run
            else:
                # Windows and 2.7 don't have _save_input
                run = subprocess.run
            if shell and isinstance(shell, six.string_types):
                executable = shell
            else:
                executable = None
            completed = run(
                # follow recommendation to use string with shell, list without
                cmdStr if shell else cmd,
                shell=shell,
                executable=executable,
                env=env,
                cwd=cwd,
                timeout=timeout,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                input=None,
            )

            # try to convert stdout and stderr to strings but leave as binary if that fails
            try:
                completed.stdout = completed.stdout.decode()
            except:
                pass
            try:
                completed.stderr = completed.stderr.decode()
            except:
                pass
            completed.cmd = cmdStr
            completed.timeout = None
            completed.error = None
            return completed
        except subprocess.TimeoutExpired as err:
            err.cmd = cmdStr
            err.timeout = timeout
            err.returncode = None
            err.error = None
            return err
        except Exception as err:
            err.cmd = cmdStr
            err.timeout = None
            err.stderr = None
            err.stdout = None
            err.returncode = None
            err.error = err
            return err

    def _handleResult(self, task, result):
        error = result.error or result.returncode or result.timeout
        if error:
            task.logger.warning("shell task run failure: %s", result.cmd)
            if result.error:
                task.logger.info("shell task error", exc_info=result.error)
            else:
                task.logger.info(
                    "shell task return code: %s, stderr: %s",
                    result.returncode,
                    _truncate(result.stderr),
                )
        else:
            task.logger.info("shell task run success: %s", result.cmd)
            task.logger.debug("shell task output: %s", _truncate(result.stdout))
        # strips terminal escapes after we printed output via logger
        result.stdout = unstyle(result.stdout or "")
        result.stderr = unstyle(result.stderr or "")
        return not error

    def _processResult(self, task, result):
        success = self._handleResult(task, result)
        resultDict = result.__dict__.copy()
        resultDict["success"] = success
        self.processResultTemplate(task, resultDict)
        if task._errors:
            return False
        return success

    def canRun(self, task):
        params = task.inputs
        cmd = params.get("command", self._defaultCmd)
        if not cmd:
            return "missing command to execute"
        if isinstance(cmd, list) and not params.get("shell") and not which(cmd[0]):
            return "'%s' is not executable" % cmd[0]
        return True

    def canDryRun(self, task):
        return task.inputs.get("dryrun")

    def run(self, task):
        params = task.inputs
        cmd = params["command"]
        isString = isinstance(cmd, six.string_types)
        # default for shell: True if command is a string otherwise False
        shell = params.get("shell", isString)
        env = task.getEnvironment(False)

        cwd = params.get("cwd")
        if cwd:
            # if cwd is relative, make it relative to task.cwd
            assert isinstance(cmd, six.string_types)
            cwd = os.path.abspath(os.path.join(task.cwd, cwd))
        else:
            cwd = task.cwd
        keeplines = params.get("keeplines")

        if task.dryRun and isinstance(task.inputs.get("dryrun"), six.string_types):
            dryrunArg = task.inputs["dryrun"]
            if "%dryrun%" in cmd:  # replace %dryrun%
                if isString:
                    cmd = cmd.replace("%dryrun%", dryrunArg)
                else:
                    cmd[cmd.index("%dryrun%")] = dryrunArg
            else:  # append dryrunArg
                if isString:
                    cmd += " " + dryrunArg
                else:
                    cmd.append(dryrunArg)
        elif "%dryrun%" in cmd:
            if isString:
                cmd = cmd.replace("%dryrun%", "")
            else:
                cmd.remove("%dryrun%")

        echo = params.get("echo")  # task.verbose > -1)
        result = self.runProcess(
            cmd,
            shell=shell,
            timeout=task.configSpec.timeout,
            env=env,
            cwd=cwd,
            keeplines=keeplines,
            echo=echo,
        )
        success = self._processResult(task, result)
        done = task.inputs.get("done", {})
        success = done.pop("success", success)
        yield task.done(success, result=result.__dict__, **done)



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
from ruamel.yaml.comments import CommentedMap
from . import __version__
from .util import getPackageDigest

class Lock(object):
    def __init__(self, ensemble):
        self.ensemble = ensemble

    def lock(self):
        """
        lock:
          runtime:
            unfurl:
              version
              digest
            toolVersions:
          repositories:
             - name:
               url:
               revision:
               initial:
               origin:
          ensembles:
            name: # e.g. localhost
              uri:
              changeId:
              digest:
              manifest:
        """
        lock = CommentedMap()
        lock["runtime"] = self.lockRuntime()
        lock["repositories"] = [
            repo.lock() for repo in self.ensemble.repositories.values()
        ]
        ensembles = self.lockEnsembles()
        if ensembles:
            lock["ensembles"] = ensembles

        # XXX artifacts should be saved as part of status because different tasks
        # could use different versions of the same artifact at different times
        # artifacts = []
        # for task in tasks:
        #     self.lockTask(task, artifacts)
        # if self.artifacts:
        #     lock["artifacts"] = self.artifacts
        return lock

    def lockRuntime(self):
        ensemble = self.ensemble
        record = CommentedMap()
        record["unfurl"] = CommentedMap(
            (("version", __version__(True)), ("digest", getPackageDigest()))
        )
        if ensemble.localEnv and ensemble.localEnv.toolVersions:
            record["toolVersions"] = {
                n: list(v) for n, v in ensemble.localEnv.toolVersions.items()
            }
        # XXX Pipfile.lock: _meta.hash, python version
        return record

    def lockEnsembles(self):
        ensemble = self.ensemble
        ensembles = CommentedMap()
        for name, _import in ensemble.imports.items():
            root = _import.resource.root.shadow or _import.resource.root
            manifest = ensemble._importedManifests.get(id(root))
            if not manifest:
                continue
            lastJob = manifest.lastJob or {}
            ensembles[name] = CommentedMap(
                [
                    ("uri", manifest.uri),
                    ("changeId", lastJob.get("changeId")),
                    ("digest", manifest.specDigest),
                    ("manifest", _import.spec["manifest"]),
                ]
            )
        return ensembles

    def lockTask(self, task, artifacts):
        instance = task.target
        artifactName = "image"
        artifact = instance.artifacts.get(artifactName)
        if artifact:
            docker_container = task.outputs.get("docker_container")
            if docker_container:
                spec = dict(
                    instance=instance.name,
                    artifact=artifactName,
                    digest=docker_container["Image"],
                    changeId=task.changeId,
                )
                path = task.outputs.get("image_path")
                if path:
                    spec["name"] = path
                artifacts.append(CommentedMap(spec.items()))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging
import six
from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError
from toscaparser.entity_template import EntityTemplate
from toscaparser.activities import Activity, ConditionClause

SECTIONS = (TARGET, TARGET_RELATIONSHIP, OPERATION_HOST, FILTER,
            ACTIVITIES, ON_SUCCESS, ON_FAILURE) = \
           ('target', 'target_relationship', 'operation_host', 'filter',
            'activities', 'on_success', 'on_failure')

log = logging.getLogger('tosca')

class Step(EntityTemplate):

    '''Step defined in workflows of topology template'''

    def __init__(self, name, step_tpl):
        self.name = name
        self.step_tpl = step_tpl
        self._validate_keys()
        self._activities = None
        self._filter = None
        for key in (TARGET, TARGET_RELATIONSHIP, OPERATION_HOST):
            setattr(self, key, self.step_tpl.get(key))
        for key in ('on_success', 'on_failure'):
            v = self.step_tpl.get(key) or []
            if isinstance(v, six.string_types):
              v = [v]
            setattr(self, key, v)
        self.activities
        self.filter

    @property
    def activities(self):
      if self._activities is None:
        self._activities = [Activity(tpl) for tpl in self.step_tpl.get(ACTIVITIES, [])]
      return self._activities

    @property
    def filter(self):
        """
        filter: # filter is a list of clauses. Matching between clauses is and.
          - or: # only one of sub-clauses must be true.
            - assert:
              - foo: [{equals: true}]
            - assert:
              - bar: [{greater_than: 2}, {less_than: 20}]
      """
        if self._filter is None:
            self._filter = list(ConditionClause.getConditions(self.step_tpl.get(FILTER, [])))
        return self._filter

    def _validate_keys(self):
        # XXX validate target and activities are present and valid
        for key in self.step_tpl.keys():
            if key not in SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='Step "%s"' % self.name,
                                      field=key))



TOSCA_VERSION = "tosca_simple_unfurl_1_0_0"

# loaded by toscaparser extention manager using entrypoint declared in setup.cfg
class plugindef_1_0_0(object):
    VERSION = TOSCA_VERSION

    DEFS_FILE = "tosca-ext.yaml"

    SECTIONS = ("metadata", "decorators")



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
"""
This module implements creating and cloning project and ensembles as well Unfurl runtimes.
"""
import uuid
import os
import os.path
import datetime
import sys
import shutil
from . import DefaultNames, getHomeConfigPath, __version__
from .repo import Repo, GitRepo, splitGitUrl, isURLorGitPath
from .util import UnfurlError
from .localenv import LocalEnv, Project
import random
import string

_templatePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "templates")


def renameForBackup(dir):
    ctime = datetime.datetime.fromtimestamp(os.stat(dir).st_ctime)
    new = dir + "." + ctime.strftime("%Y-%m-%d-%H-%M-%S")
    os.rename(dir, new)
    return new


def get_random_password(count=12, prefix="uv"):
    srandom = random.SystemRandom()
    start = string.ascii_letters + string.digits
    source = string.ascii_letters + string.digits + "%&()*+,-./:<>?=@^_`~"
    return prefix + "".join(
        srandom.choice(source if i else start) for i in range(count)
    )


def _writeFile(folder, filename, content):
    if not os.path.isdir(folder):
        os.makedirs(os.path.normpath(folder))
    filepath = os.path.join(folder, filename)
    with open(filepath, "w") as f:
        f.write(content)
    return filepath


def writeTemplate(folder, filename, template, vars, templateDir=None):
    from .runtime import NodeInstance
    from .eval import RefContext
    from .support import applyTemplate

    if templateDir and not os.path.isabs(templateDir):
        templateDir = os.path.join(_templatePath, templateDir)
    if not templateDir or not os.path.exists(os.path.join(templateDir, template)):
        templateDir = _templatePath  # use default file if missing from templateDir

    with open(os.path.join(templateDir, template)) as f:
        source = f.read()
    instance = NodeInstance()
    instance._baseDir = _templatePath
    content = applyTemplate(source, RefContext(instance, vars))
    return _writeFile(folder, filename, content)


def writeProjectConfig(
    projectdir,
    filename=DefaultNames.LocalConfig,
    templatePath=DefaultNames.LocalConfig + ".j2",
    vars=None,
    templateDir=None,
):
    _vars = dict(include="", manifestPath=None)
    if vars:
        _vars.update(vars)
    return writeTemplate(projectdir, filename, templatePath, _vars, templateDir)


def createHome(home=None, render=False, replace=False, **kw):
    """
    Create the home project if missing
    """
    homePath = getHomeConfigPath(home)
    if not homePath:
        return None
    exists = os.path.exists(homePath)
    if exists and not replace:
        return None

    homedir, filename = os.path.split(homePath)
    if render:  # just render
        repo = Repo.findContainingRepo(homedir)
        # XXX if repo and update: git stash; git checkout rendered
        ensembleDir = os.path.join(homedir, DefaultNames.EnsembleDirectory)
        ensembleRepo = Repo.findContainingRepo(ensembleDir)
        configPath = renderProject(homedir, repo, ensembleRepo, "home")
        # XXX if repo and update: git commit -m"updated"; git checkout master; git stash pop
        return configPath
    else:
        if exists:
            renameForBackup(homedir)

    newHome, configPath, repo = createProject(
        homedir,
        template="home",
        runtime=kw.get("runtime") or "venv:",
        no_runtime=kw.get("no_runtime"),
        msg="Create the unfurl home repository",
    )
    if repo:
        repo.repo.git.branch("rendered")  # now create a branch
    return configPath


def _createRepo(gitDir, ignore=True):
    import git

    if not os.path.isdir(gitDir):
        os.makedirs(gitDir)
    repo = git.Repo.init(gitDir)
    repo.index.add(addHiddenGitFiles(gitDir))
    repo.index.commit("Initial Commit for %s" % uuid.uuid1())

    if ignore:
        Repo.ignoreDir(gitDir)
    return GitRepo(repo)


def writeServiceTemplate(projectdir):
    from .tosca import TOSCA_VERSION

    vars = dict(version=TOSCA_VERSION)
    return writeTemplate(
        projectdir, "service-template.yaml", "service-template.yaml.j2", vars
    )


def writeEnsembleManifest(
    destDir, manifestName, specRepo, specDir=None, extraVars=None, templateDir=None
):
    if specDir:
        specDir = os.path.abspath(specDir)
    else:
        specDir = ""
    vars = dict(specRepoUrl=specRepo.getGitLocalUrl(specDir, "spec"))
    if extraVars:
        vars.update(extraVars)
    return writeTemplate(destDir, manifestName, "manifest.yaml.j2", vars, templateDir)


def addHiddenGitFiles(gitDir):
    # write .gitignore and  .gitattributes
    gitIgnorePath = writeTemplate(gitDir, ".gitignore", "gitignore.j2", {})
    gitAttributesContent = "**/*%s merge=union\n" % DefaultNames.JobsLog
    gitAttributesPath = _writeFile(gitDir, ".gitattributes", gitAttributesContent)
    return [os.path.abspath(gitIgnorePath), os.path.abspath(gitAttributesPath)]


def renderProject(
    projectdir,
    repo,
    ensembleRepo,
    templateDir=None,
    names=DefaultNames,
):
    """
    Creates a folder named `projectdir` with a git repository with the following files:

    unfurl.yaml
    local/unfurl.yaml
    ensemble-template.yaml
    ensemble/ensemble.yaml

    Returns the absolute path to unfurl.yaml
    """
    assert os.path.isabs(projectdir), projectdir + " must be an absolute path"
    # write the project files
    localConfigFilename = names.LocalConfig

    vars = dict(vaultpass=get_random_password())
    writeProjectConfig(
        os.path.join(projectdir, "local"),
        localConfigFilename,
        "unfurl.local.yaml.j2",
        vars,
        templateDir,
    )

    localInclude = "+?include: " + os.path.join("local", localConfigFilename)
    vars = dict(include=localInclude)
    projectConfigPath = writeProjectConfig(
        projectdir,
        names.LocalConfig,
        "unfurl.yaml.j2",
        vars,
        templateDir,
    )

    # write ensemble-template.yaml
    writeTemplate(
        projectdir,
        names.EnsembleTemplate,
        "manifest-template.yaml.j2",
        {},
        templateDir,
    )

    if ensembleRepo:
        ensembleDir = os.path.join(projectdir, names.EnsembleDirectory)
        manifestName = names.Ensemble
        extraVars = dict(
            ensembleUri=ensembleRepo.getUrlWithPath(
                os.path.join(ensembleDir, manifestName)
            ),
            # include the ensembleTemplate in the root of the specDir
            ensembleTemplate=names.EnsembleTemplate,
        )
        # write ensemble/ensemble.yaml
        writeEnsembleManifest(
            ensembleDir,
            manifestName,
            repo,
            extraVars=extraVars,
            templateDir=templateDir,
        )
    return projectConfigPath


def createProject(
    projectdir,
    home=None,
    mono=False,
    existing=False,
    empty=False,
    template=None,
    submodule=False,
    **kw
):
    if existing:
        repo = Repo.findContainingRepo(projectdir)
        if not repo:
            raise UnfurlError("Could not find an existing repository")
    else:
        repo = None

    newHome = ""
    homePath = getHomeConfigPath(home)
    # don't try to create the home project if we are already creating the home project
    if projectdir != os.path.dirname(homePath):
        # create the home project (but only if it doesn't exist already)
        newHome = createHome(home, **kw)

    if repo:
        addHiddenGitFiles(projectdir)
    else:
        repo = _createRepo(projectdir)

    ensembleDir = os.path.join(projectdir, DefaultNames.EnsembleDirectory)
    if mono:
        ensembleRepo = repo
    else:
        ensembleRepo = _createRepo(ensembleDir, not submodule)

    projectConfigPath = renderProject(
        projectdir,
        repo,
        not empty and ensembleRepo,
        template,
    )
    if not mono:
        ensembleRepo.addAll(ensembleDir)
        ensembleRepo.repo.index.commit("Default ensemble repository boilerplate")

    if submodule:
        repo.addSubModule(ensembleDir)

    if not newHome and not kw.get("no_runtime") and kw.get("runtime"):
        # if runtime was explicitly set and we aren't creating the home project
        # then initialize the runtime here
        try:
            initEngine(projectdir, kw.get("runtime"))
        except:
            pass  # don't stop even if this fails

    repo.addAll(projectdir)
    repo.repo.index.commit(kw.get("msg") or "Create a new Unfurl project")

    return newHome, projectConfigPath, repo


def cloneLocalRepos(manifest, sourceProject, targetProject):
    # We need to clone repositories that are local to the source project
    # otherwise we won't be able to find them
    for repoView in manifest.repositories.values():
        repoSpec = repoView.repository
        if repoSpec.name == "self":
            continue
        repo = sourceProject.findGitRepoFromRepository(repoSpec)
        if repo:
            targetProject.findOrClone(repo)


def _createEnsembleRepo(manifest, repo):
    destDir = os.path.dirname(manifest.manifest.path)
    if not repo:
        repo = _createRepo(destDir)
    elif not os.path.isdir(destDir):
        os.makedirs(destDir)

    manifest.metadata["uri"] = repo.getUrlWithPath(manifest.manifest.path)
    with open(manifest.manifest.path, "w") as f:
        manifest.dump(f)

    repo.repo.index.add([manifest.manifest.path])
    repo.repo.index.commit("Default ensemble repository boilerplate")
    return repo


def _looksLike(path, name):
    # in case path is a directory:
    if os.path.isfile(os.path.join(path, name)):
        return path, name
    if path.endswith(name):  # name is explicit so don't need to check if file exists
        return os.path.split(path)
    return None


def _getEnsemblePaths(sourcePath, sourceProject):
    """
    Returns either a pointer to the ensemble to clone
    or a dict of variables to pass to an ensemble template to create a new one

    if sourcePath doesn't exist, return {}

    look for an ensemble given sourcePath (unless sourcePath looks like a service template)
    if that fails look for (ensemble-template, service-template) if sourcePath is a directory
    otherwise
        return {}
    """
    template = None
    if not os.path.exists(sourcePath or "."):
        return {}
    isServiceTemplate = sourcePath.endswith(DefaultNames.ServiceTemplate)
    if not isServiceTemplate:
        # we only support cloning TOSCA service templates if their names end in "service-template.yaml"
        try:
            localEnv = LocalEnv(sourcePath, project=sourceProject)
            sourceDir = sourceProject.getRelativePath(
                os.path.dirname(localEnv.manifestPath)
            )
            return dict(sourceDir=sourceDir, localEnv=localEnv)
        except:
            pass

    # didn't find the specified file (or the default ensemble if none was specified)
    # so if sourcePath was a directory try for one of the default template files
    if isServiceTemplate or os.path.isdir(sourcePath):
        # look for an ensemble-template or service-template in source path
        template = _looksLike(sourcePath, DefaultNames.EnsembleTemplate)
        if template:
            sourceDir = sourceProject.getRelativePath(template[0])
            return dict(sourceDir=sourceDir, ensembleTemplate=template[1])
        template = _looksLike(sourcePath, DefaultNames.ServiceTemplate)
        if template:
            sourceDir = sourceProject.getRelativePath(template[0])
            return dict(sourceDir=sourceDir, serviceTemplate=template[1])
        # nothing valid found
    return {}


def createNewEnsemble(templateVars, project, targetPath, mono):
    """
    If "localEnv" is in templateVars, clone that ensemble;
    otherwise create one from a template with templateVars
    """
    # targetPath is relative to the project root
    from unfurl import yamlmanifest

    assert not os.path.isabs(targetPath)
    if not targetPath or targetPath == ".":
        destDir, manifestName = DefaultNames.EnsembleDirectory, DefaultNames.Ensemble
    elif targetPath.endswith(".yaml") or targetPath.endswith(".yml"):
        destDir, manifestName = os.path.split(targetPath)
    else:
        destDir = targetPath
        manifestName = DefaultNames.Ensemble
    # choose a destDir that doesn't conflict with an existing folder
    # (i.e. if default ensemble already exists)
    destDir = project.getUniquePath(destDir)
    # destDir is now absolute
    targetPath = os.path.normpath(os.path.join(destDir, manifestName))

    if "localEnv" not in templateVars:
        # we found a template file to clone
        assert project
        sourceDir = os.path.normpath(
            os.path.join(project.projectRoot, templateVars["sourceDir"])
        )
        specRepo, relPath, revision, bare = project.findPathInRepos(sourceDir)
        if not specRepo:
            raise UnfurlError(
                '"%s" is not in a git repository. Cloning from plain file directories not yet supported'
                % os.path.abspath(sourceDir)
            )
        manifestPath = writeEnsembleManifest(
            os.path.join(project.projectRoot, destDir),
            manifestName,
            specRepo,
            sourceDir,
            templateVars,
        )
        localEnv = LocalEnv(manifestPath, project=project)
        manifest = yamlmanifest.ReadOnlyManifest(localEnv=localEnv)
    elif templateVars:
        # didn't find a template file
        # look for an ensemble at the given path or use the source project's default
        manifest = yamlmanifest.clone(templateVars["localEnv"], targetPath)
    else:
        raise UnfurlError("can't find anything to clone")
    _createEnsembleRepo(manifest, mono and project.projectRepo)
    return destDir, manifest
    # XXX need to add manifest to unfurl.yaml


def cloneRemoteProject(source, destDir):
    # check if source is a git url
    repoURL, filePath, revision = splitGitUrl(source)
    # not yet supported: add repo to local project
    # destRoot = Project.findPath(destDir)
    # if destRoot:
    #     # destination is in an existing project, use that one
    #     sourceProject = Project(destRoot)
    #     repo = sourceProject.findGitRepo(repoURL, revision)
    #     if not repo:
    #         repo = sourceProject.createWorkingDir(repoURL, revision)
    #     source = os.path.join(repo.workingDir, filePath)
    # else: # otherwise clone to dest
    if os.path.exists(destDir) and os.listdir(destDir):
        raise UnfurlError(
            'Can not clone project into "%s": folder is not empty' % destDir
        )
    Repo.createWorkingDir(repoURL, destDir, revision)  # clone source repo
    targetDir = os.path.join(destDir, filePath)
    sourceRoot = Project.findPath(targetDir)
    if not sourceRoot:
        raise UnfurlError(
            'Error: cloned "%s" to "%s" but couldn\'t find an Unfurl project'
            % (source, destDir)
        )
    sourceProject = Project(sourceRoot)
    return sourceProject, targetDir


def getSourceProject(source):
    sourceRoot = Project.findPath(source)
    if sourceRoot:
        return Project(sourceRoot)
    return None


def isEnsembleInProjectRepo(project, paths):
    # check if source points to an ensemble that is part of the project repo
    if not project.projectRepo or "localEnv" not in paths:
        return False
    sourceDir = paths["sourceDir"]
    assert not os.path.isabs(sourceDir)
    pathToEnsemble = os.path.join(project.projectRoot, sourceDir)
    if not os.path.isdir(pathToEnsemble):
        return False
    if project.projectRepo.isPathExcluded(sourceDir):
        return False
    return True


def clone(source, dest, includeLocal=False, **options):
    """
    Clone the `source` ensemble to `dest`. If `dest` isn't in a project, create one.
    `source` can point to an ensemble_template, a service_template, an existing ensemble
    or a folder containing one of those. If it points to a project its default ensemble will be cloned.

    Referenced `repositories` will be cloned if a git repository or copied if a regular file folder,
    If the folders already exist they will be copied to new folder unless the git repositories have the same HEAD.
    but the local repository names will remain the same.

    ================ =============================================
    dest             result
    ================ =============================================
    Inside project   new ensemble
    new or empty dir clone or create project (depending on source)
    another project  error (not yet supported)
    other            error
    ================ =============================================

    """
    if not dest:
        dest = Repo.getPathForGitRepo(source)  # choose dest based on source url
    # XXX else: # we're assuming dest is directory, handle case where filename is included

    mono = options.get("mono") or options.get("existing")
    isRemote = isURLorGitPath(source)
    if isRemote:
        clonedProject, source = cloneRemoteProject(source, dest)
        # source is now a path inside the cloned project
        paths = _getEnsemblePaths(source, clonedProject)
    else:
        sourceProject = getSourceProject(source)
        sourceNotInProject = not sourceProject
        if sourceNotInProject:
            # source wasn't in a project
            raise UnfurlError(
                "Can't clone \"%s\": it isn't in an Unfurl project or repository"
                % source
            )
            # XXX create a new project from scratch for the ensemble
            # if os.path.exists(dest) and os.listdir(dest):
            #     raise UnfurlError(
            #         'Can not create a project in "%s": folder is not empty' % dest
            #     )
            # newHome, projectConfigPath, repo = createProject(
            #     dest, emtpy=True, **options
            # )
            # sourceProject = Project(projectConfigPath)

        relDestDir = sourceProject.getRelativePath(dest)
        paths = _getEnsemblePaths(source, sourceProject)
        if not relDestDir.startswith(".."):
            # dest is in the source project (or its a new project)
            # so don't need to clone, just need to create an ensemble
            destDir, manifest = createNewEnsemble(
                paths, sourceProject, relDestDir, mono
            )
            return 'Created ensemble in %s project: "%s"' % (
                "new" if sourceNotInProject else "existing",
                os.path.abspath(destDir),
            )
        else:
            # XXX we are not trying to adjust the clone location to be a parent of dest
            sourceProject.projectRepo.clone(dest)
            relPathToProject = sourceProject.projectRepo.findRepoPath(
                sourceProject.projectRoot
            )
            # adjust if project is not at the root of its repo:
            dest = Project.normalizePath(os.path.join(dest, relPathToProject))
            clonedProject = Project(dest)

    # pass in "" as dest because we already "consumed" dest by cloning the project to that location
    manifest, message = _createInClonedProject(paths, clonedProject, "", mono)
    if not isRemote and manifest:
        # we need to clone referenced local repos so the new project has access to them
        cloneLocalRepos(manifest, sourceProject, clonedProject)
    return message


def _createInClonedProject(paths, clonedProject, dest, mono):
    """
    Called by `clone` when cloning an ensemble.

    ================================   ========================
    source ensemble                    result
    ================================   ========================
    project root or ensemble in repo   git clone only
    local ensemble or template         git clone + new ensemble
    ================================   ========================

    """
    from unfurl import yamlmanifest

    ensembleInProjectRepo = isEnsembleInProjectRepo(clonedProject, paths)
    if ensembleInProjectRepo:
        # the ensemble is already part of the source project repository or a submodule
        # we're done
        manifest = yamlmanifest.ReadOnlyManifest(localEnv=paths["localEnv"])
        return manifest, "Cloned project to " + clonedProject.projectRoot
    else:
        # create local/unfurl.yaml in the new project
        # XXX vaultpass should only be set for the new ensemble being created
        writeProjectConfig(
            os.path.join(clonedProject.projectRoot, "local"),
            DefaultNames.LocalConfig,
            "unfurl.local.yaml.j2",
            dict(vaultpass=get_random_password()),
        )
        # dest: should be a path relative to the clonedProject's root
        assert not os.path.isabs(dest)
        destDir, manifest = createNewEnsemble(paths, clonedProject, dest, mono)
        return manifest, 'Created new ensemble at "%s" in cloned project at "%s"' % (
            destDir,
            clonedProject.projectRoot,
        )


def _getUnfurlRequirementUrl(spec):
    """Expand the given string in an URL for installing the local Unfurl package.

    If @ref is omitted the tag for the current release will be used,
    if empty ("@") the latest revision will be used
    If no path or url is specified https://github.com/onecommons/unfurl.git will be used.

    For example:

    @tag
    ./path/to/local/repo
    ./path/to/local/repo@tag
    ./path/to/local/repo@
    git+https://example.com/forked/unfurl.git
    @

    Args:
        spec (str): can be a path to a git repo, git url or just a revision or tag.

    Returns:
      str: Description of returned object.

    """
    if not spec:
        return spec
    if "egg=unfurl" in spec:
        # looks fully specified, just return it
        return spec

    url, sep, ref = spec.rpartition("@")
    if sep:
        if ref:
            ref = "@" + ref
    else:
        ref = "@" + __version__()

    if not url:
        return "git+https://github.com/onecommons/unfurl.git" + ref + "#egg=unfurl"
    if not url.startswith("git+"):
        return "git+file://" + os.path.abspath(url) + ref + "#egg=unfurl"
    else:
        return url + ref + "#egg=unfurl"


def initEngine(projectDir, runtime):
    runtime = runtime or "venv:"
    kind, sep, rest = runtime.partition(":")
    if kind == "venv":
        pipfileLocation, sep, unfurlLocation = rest.partition(":")
        return createVenv(
            projectDir, pipfileLocation, _getUnfurlRequirementUrl(unfurlLocation)
        )
    # elif kind == 'docker'
    return "unrecognized runtime uri"


def _runPipEnv(do_install, kw):
    # create the virtualenv and install the dependencies specified in the Pipefiles
    sys_exit = sys.exit
    try:
        retcode = 0

        def noexit(code):
            retcode = code

        sys.exit = noexit

        do_install(**kw)
    finally:
        sys.exit = sys_exit

    return retcode


def createVenv(projectDir, pipfileLocation, unfurlLocation):
    """Create a virtual python environment for the given project."""

    os.environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
    VIRTUAL_ENV = os.environ.get("VIRTUAL_ENV")
    os.environ["PIPENV_VENV_IN_PROJECT"] = "1"
    if "PIPENV_PYTHON" not in os.environ:
        os.environ["PIPENV_PYTHON"] = sys.executable

    if pipfileLocation:
        pipfileLocation = os.path.abspath(pipfileLocation)

    try:
        cwd = os.getcwd()
        os.chdir(projectDir)
        # need to set env vars and change current dir before importing pipenv
        from pipenv.core import do_install, ensure_python
        from pipenv.utils import python_version
        from pipenv import environments

        pythonPath = str(ensure_python())
        assert pythonPath, pythonPath
        if not pipfileLocation:
            versionStr = python_version(pythonPath)
            assert versionStr, versionStr
            version = versionStr.rpartition(".")[0]  # 3.8.1 => 3.8
            # version = subprocess.run([pythonPath, "-V"]).stdout.decode()[
            #     7:10
            # ]  # e.g. Python 3.8.1 => 3.8
            pipfileLocation = os.path.join(
                _templatePath, "python" + version
            )  # e.g. templates/python3.8

        if not os.path.isdir(pipfileLocation):
            return 'Pipfile location is not a valid directory: "%s"' % pipfileLocation

        # copy Pipfiles to project root
        if os.path.abspath(projectDir) != os.path.abspath(pipfileLocation):
            for filename in ["Pipfile", "Pipfile.lock"]:
                path = os.path.join(pipfileLocation, filename)
                if os.path.isfile(path):
                    shutil.copy(path, projectDir)

        kw = dict(python=pythonPath)
        # need to run without args first so lock isn't overwritten
        retcode = _runPipEnv(do_install, kw)
        if retcode:
            return "Pipenv (step 1) failed: %s" % retcode

        # we need to set these so pipenv doesn't try to recreate the virtual environment
        environments.PIPENV_USE_SYSTEM = 1
        environments.PIPENV_IGNORE_VIRTUALENVS = False
        os.environ["VIRTUAL_ENV"] = os.path.join(projectDir, ".venv")
        environments.PIPENV_VIRTUALENV = os.path.join(projectDir, ".venv")

        # we need to set skip_lock or pipenv will not honor the existing lock
        kw["skip_lock"] = True
        if unfurlLocation:
            kw["editable_packages"] = [unfurlLocation]
        else:
            kw["packages"] = [
                "unfurl==" + __version__()
            ]  # use the same version as current
        retcode = _runPipEnv(do_install, kw)
        if retcode:
            return "Pipenv (step 2) failed: %s" % retcode

        return ""
    finally:
        if VIRTUAL_ENV:
            os.environ["VIRTUAL_ENV"] = VIRTUAL_ENV
        else:
            os.environ.pop("VIRTUAL_ENV", None)
        os.chdir(cwd)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging

from toscaparser.common.exception import ExceptionCollector
from toscaparser.common.exception import UnknownFieldError, ValidationError
from toscaparser.entity_template import EntityTemplate
from toscaparser.elements.constraints import Constraint, get_constraint_class


SECTIONS = (DELEGATE, SET_STATE, CALL_OPERATION, INLINE) = \
              ('delegate', 'set_state', 'call_operation', 'inline')

log = logging.getLogger('tosca')

PRECONDITION_SECTIONS = (TARGET, TARGET_RELATIONSHIP, CONDITION) = (
                          'target', 'target_relationship', "condition")

class Precondition(EntityTemplate):

    '''Step defined in workflows of topology template'''

    def __init__(self, precondition_tpl):
        self.precondition_tpl = precondition_tpl
        self._validate_keys()
        self._condition = None
        self.target = self.precondition_tpl.get('target')
        self.target_relationship = self.precondition_tpl.get('target_relationship')

    def _validate_keys(self):
        for key in self.precondition_tpl.keys():
            if key not in PRECONDITION_SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='PreCondition',
                                      field=key))

    @property
    def condition(self):
        """
        filter: # filter is a list of clauses. Matching between clauses is and.
          - or: # only one of sub-clauses must be true.
            - assert:
              - foo: [{equals: true}]
            - assert:
              - bar: [{greater_than: 2}, {less_than: 20}]
      """
        if self._condition is None:
            self._condition = list(ConditionClause.getConditions(self.precondition_tpl.get(CONDITION, [])))
        return self._condition

class ConditionClause(object):

  def __init__(self, name, definition):
    self.name = name
    if name in ['and', 'or', 'not', 'assert']:
        self.conditions = list(self.getConditions(definition))
    else: # 3.6.25.3 Direct assertion definition
        if isinstance(definition, dict):
            definition = [definition]
        # hack: pick a prop_type to avoid validation error
        # XXX need to get the real property_type from the target's attribute definition
        self.conditions = [Constraint(name,
                            get_constraint_class(next(iter(constraint))).valid_prop_types[0],
                            constraint) for constraint in definition]

  def evaluate(self, attributes):
    if self.name == 'not':
        for condition in self.conditions:
            if not condition.evaluate(attributes):
                return True
        return False
    elif self.name in ['and', 'assert']:
        for condition in self.conditions:
            if not condition.evaluate(attributes):
                return False
        return True
    elif self.name == 'or':
        for condition in self.conditions:
            if condition.evaluate(attributes):
                return True
    else:
        # 3.6.25.3 Direct assertion definition
        if self.name not in attributes:
            return False
        value = attributes[self.name]
        for condition in self.conditions:
            # XXX:
            #if condition.property_type in scalarunit.ScalarUnit.SCALAR_UNIT_TYPES:
            #  value = scalarunit.get_scalarunit_value(condition.property_type, value)
            if not condition._is_valid(value):
                return False
        return True

  @staticmethod
  def getConditions(constraints):
      for constraint in constraints:
          key, value = list(constraint.items())[0]
          yield ConditionClause(key, value)

class Activity(object):

    '''Step defined in workflows of topology template'''

    def __init__(self, activity_tpl):
        self.activity_tpl = activity_tpl
        self._validate_keys()
        self.type = list(self.activity_tpl.keys())[0]
        val = self.activity_tpl[self.type]
        if isinstance(val, dict):
            self.inputs = val.pop('inputs', None)
            # the other item will be a string:
            # delegate and inline both take workflow and inputs, call_operation takes operation and inputs
            setattr(self, self.type, list(val.values())[0])
        else:
            setattr(self, self.type, val)
            self.inputs = None

    def _validate_keys(self):
        if len(self.activity_tpl) != 1:
            ExceptionCollector.appendException(
                ValidationError(message="Invalid Activity"))

        for key in self.activity_tpl.keys():
            if key not in SECTIONS:
                ExceptionCollector.appendException(
                    UnknownFieldError(what='Activity',
                                      field=key))



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
import six
from ..configurator import Configurator, Status
from ..result import ResultsMap
from ..util import registerShortNames

# need to define these now because because these configurators are lazily imported
# and so won't register themselves through AutoRegisterClass
registerShortNames(
    {
        name: "unfurl.configurators.%s.%sConfigurator" % (name.lower(), name)
        for name in "Ansible Shell Supervisor Terraform".split()
    }
)


class TemplateConfigurator(Configurator):
    def processResultTemplate(self, task, result):
        """
        for both the ansible and shell configurators
        result can include: "returncode", "msg", "error", "stdout", "stderr"
        Ansible also includes "outputs"
        """
        # get the resultTemplate without evaluating it
        resultTemplate = task.inputs._attributes.get("resultTemplate")
        if resultTemplate:  # evaluate it now with the result
            if isinstance(resultTemplate, six.string_types):
                query = dict(template=resultTemplate)
            else:
                query = resultTemplate

            # workaround for jinja template processing setting Result when getting items
            if not isinstance(result, ResultsMap):
                vars = ResultsMap(result, task.inputs.context)
                vars.doFullResolve = True
            else:
                vars = result
            results = task.query({"eval": query}, vars=vars)
            if results:
                task.updateResources(results)

    def canDryRun(self, task):
        return not not task.inputs.get("dryrun")

    def run(self, task):
        if task.dryRun:
            runResult = task.inputs.get("dryrun")
            if not isinstance(runResult, dict):
                runResult = task.inputs.get("run")
        else:
            runResult = task.inputs.get("run")

        done = task.inputs.get("done", {})
        if "result" not in done:
            if not isinstance(runResult, dict):
                done["result"] = {"run": runResult}
            else:
                done["result"] = runResult
        self.processResultTemplate(task, done.get("result"))
        yield task.done(**done)


class DelegateConfigurator(Configurator):
    def canDryRun(self, task):
        return True  # ok because this will also be called on the subtask

    def run(self, task):
        subtaskRequest = task.createSubTask(
            task.inputs["operation"], task.inputs.get("target")
        )
        assert subtaskRequest
        # note: this will call canRun() and if needed canDryRun() on subtask but not shouldRun()
        subtask = yield subtaskRequest
        yield subtask.result



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
import codecs
from ..configurator import Configurator, Status
from ..runtime import RelationshipInstance
from .ansible import AnsibleConfigurator
import json
from ansible.module_utils.k8s.common import K8sAnsibleMixin


def _getConnectionConfig(instance):
    # see https://docs.ansible.com/ansible/latest/modules/k8s_module.html#k8s-module
    #  for connection settings
    if not instance:
        return {}
    connect = {}
    if isinstance(instance, RelationshipInstance):
        connect = instance.attributes
        endpoint = instance.parent.attributes
    else:
        endpoint = instance.attributes

    connection = {}
    if "host" in endpoint:  # endpoint capability only
        connection["host"] = endpoint["host"]

    map1 = {"KUBECONFIG": "kubeconfig", "context": "context", "secure": "verify_ssl"}
    # relationship overrides capability
    for attributes in [endpoint, connect]:
        for key, value in map1.items():
            if key in attributes:
                connection[value] = attributes[key]

        credential = attributes.get("credential")
        if credential:
            if credential.get("token_type") in ["api_key", "password"]:
                connection[credential["token_type"]] = credential["token"]
            if "user" in credential:
                connection["username"] = credential["user"]
            if "keys" in credential:
                # ["ssl_ca_cert", "cert_file", "key_file"]
                connection.update(credential["keys"])

    return connection


def _getConnection(task, cluster):
    instance = task.findConnection(
        cluster, relation="unfurl.relationships.ConnectsTo.K8sCluster"
    )
    return _getConnectionConfig(instance)


class ClusterConfigurator(Configurator):
    @staticmethod
    def _getHost(connectionConfig):
        client = K8sAnsibleMixin().get_api_client(**connectionConfig)
        return client.configuration.host

    def canRun(self, task):
        if task.configSpec.operation not in ["check", "discover"]:
            return "Configurator can't perform this operation (only supports check and discover)"
        return True

    def run(self, task):
        cluster = task.target
        connectionConfig = _getConnection(task, cluster)
        try:
            cluster.attributes["apiServer"] = self._getHost(connectionConfig)
        except:
            yield task.done(
                False,
                captureException="error while trying to establish connection to cluster",
            )
        else:
            # we aren't modifying this cluster but we do want to assert that its ok
            yield task.done(True, False, Status.ok)


class ResourceConfigurator(AnsibleConfigurator):
    def getGenerator(self, task):
        if task.dryRun:
            return self.dryRun(task)
        else:
            return self.run(task)

    def dryRun(self, task):
        # XXX don't use print()
        print("generating playbook")
        # print(self.findPlaybook(task))
        # print(self.findPlaybook(task))
        print(json.dumps(self.findPlaybook(task), indent=4))
        yield task.done(True)

    def _getConnection(self, task):
        # get the cluster that the target resource is hosted on
        cluster = task.query("[.type=unfurl.nodes.K8sCluster]")
        if not cluster:
            return {}
        return _getConnection(task, cluster)

    def makeSecret(self, data):
        # base64 adds trailing \n so strip it out
        return dict(
            type="Opaque",
            apiVersion="v1",
            kind="Secret",
            data={
                k: codecs.encode(str(v).encode(), "base64").decode().strip()
                for k, v in data.items()
            },
        )

    def getDefinition(self, task):
        if task.target.template.isCompatibleType("unfurl.nodes.K8sNamespace"):
            return dict(apiVersion="v1", kind="Namespace")

        if "definition" in task.target.attributes:
            definition = task.target.attributes.getCopy("definition")
        else:
            definition = task.target.attributes.getCopy("apiResource", {})

        if not definition and task.target.template.isCompatibleType(
            "unfurl.nodes.K8sSecretResource"
        ):
            return self.makeSecret(task.target.attributes.get("data", {}))
        else:
            # XXX if definition is string: parse
            # get copy so subsequent modifications dont affect the definition
            return definition

    def updateMetadata(self, definition, task):
        namespace = None
        if task.target.parent.template.isCompatibleType("unfurl.nodes.K8sNamespace"):
            namespace = task.target.parent.attributes["name"]
        md = definition.setdefault("metadata", {})
        if namespace and "namespace" not in md:
            md["namespace"] = namespace
        # else: error if namespace mismatch?

        # XXX if using target.name, convert into kube friendly dns-style name
        name = task.target.attributes.get("name", task.target.name)
        if "name" in md and md["name"] != name:
            task.target.attributes["name"] = md["name"]
        else:
            md["name"] = name

    def findPlaybook(self, task):
        definition = self.getDefinition(task)
        self.updateMetadata(definition, task)
        delete = task.configSpec.operation in ["Standard.delete", "delete"]
        state = "absent" if delete else "present"
        connectionConfig = self._getConnection(task)
        moduleSpec = dict(state=state, **connectionConfig)
        if task.configSpec.operation in ["check", "discover"]:
            moduleSpec["kind"] = definition.get("kind", "")
            moduleSpec["name"] = definition["metadata"]["name"]
            if "namespace" in definition["metadata"]:
                moduleSpec["namespace"] = definition["metadata"]["namespace"]
        else:
            moduleSpec["resource_definition"] = definition
        return [dict(k8s=moduleSpec)]

    def processResult(self, task, result):
        # overrides super.processResult
        resource = result.result.get("result")
        task.target.attributes["apiResource"] = resource
        if resource:
            data = resource.get("kind") == "Secret" and resource.get("data")
            if data:
                resource["data"] = {k: task.sensitive(v) for k, v in data.items()}
            if task.configSpec.operation in ["check", "discover"]:
                states = dict(
                    Active=Status.ok,
                    Terminating=Status.absent,
                    Pending=Status.pending,
                    Running=Status.ok,
                    Succeeded=Status.absent,
                    Failed=Status.error,
                    Unknown=Status.unknown,
                )
                status = resource.get("status", {}).get("phase", "Unknown")
                result.status = states.get(status, Status.unknown)
        return result

    def getResultKeys(self, task, results):
        # save first time even if it hasn't changed
        return ["result"]  # also "method", "diff", invocation



import unittest
import os
import json
from unfurl.result import ResultsList, serializeValue
from unfurl.eval import Ref, mapValue, RefContext
from unfurl.support import applyTemplate
from unfurl.util import sensitive_str
from unfurl.runtime import NodeInstance
from ruamel.yaml.comments import CommentedMap


class EvalTest(unittest.TestCase):
    longMessage = True

    def test_CommentedMap(self):
        cm = CommentedMap()
        # check bug in ruamel.yaml is fixed: raises TypeError: source has undefined order
        self.assertEqual(cm, cm.copy())

    def test_CommentedMapEquality(self):
        cm = CommentedMap((("b", 2),))
        cm.insert(1, "a", 1, comment="a comment")
        self.assertEqual(cm, {"a": 1, "b": 2})

    def _getTestResource(self, more=None):
        resourceDef = {
            "name": "test",
            "a": {"ref": "name"},
            "b": [1, 2, 3],
            "d": {"a": "va", "b": "vb"},
            "n": {"n": {"n": "n"}},
            "s": {"ref": "."},
            "x": [
                {
                    "a": [
                        {"c": 1},
                        {"c": 2},
                        {"b": "exists"},
                        {"l": ["l1"]},
                        {"l": ["l2"]},
                    ]
                },
                [{"c": 5}],
                {"a": [{"c": 3}, {"c": 4}, {"l": ["l3"]}, {"l": ["l4"]}]},
                [{"c": 6}],
            ],
            "e": {"a1": {"b1": "v1"}, "a2": {"b2": "v2"}},
            "f": {"a": 1, "b": {"ref": ".::f::a"}},
        }
        if more:
            resourceDef.update(more)
        resource = NodeInstance("test", attributes=resourceDef)
        assert resource.attributes["x"] == resourceDef["x"]
        assert resource.attributes["a"] == "test"
        assert resource.attributes["s"] is resource
        return resource

    def test_refs(self):
        assert Ref.isRef({"ref": "::name"})
        assert not Ref.isRef({"ref": "::name", "somethingUnexpected": 1})
        assert Ref.isRef({"ref": "::name", "vars": {"a": None}})

    def test_refPaths(self):
        resource = self._getTestResource()
        for (exp, expected) in [
            ["x?::a[c=4]", [[{"c": 3}, {"c": 4}, {"l": ["l3"]}, {"l": ["l4"]}]]],
            [
                "x::a[c]?",
                [[{"c": 1}, {"c": 2}, {"b": "exists"}, {"l": ["l1"]}, {"l": ["l2"]}]],
            ],
            ["x::a::[c]", [{"c": 1}, {"c": 2}, {"c": 3}, {"c": 4}]],
            ["x::a?::[c]", [{"c": 1}, {"c": 2}]],
            ["a", ["test"]],
            ["b", [[1, 2, 3]]],
            ["b::0", [1]],
            ["b?::2", [3]],
            ["[b::2]::b::2", [3]],
            ["b::1", [2]],
            ["s::b", [[1, 2, 3]]],
            ["s::b::1", [2]],
            ["s::s::b::1", [2]],
            ["n::n::n", ["n"]],
            ["d[a=va]", [{"a": "va", "b": "vb"}]],
            ["d[a=vb]", []],
            ["b[1=2]", [[1, 2, 3]]],
            ["b[1=1]", []],
            ["a[=test]", ["test"]],
            ["a[!=test]", []],
            ["a[key]", []],
            ["d[a=va][b=vb]", [{"a": "va", "b": "vb"}]],
            ["d[a=va][a=vb]", []],
            ["d[a=va][a!=vb]", [{"a": "va", "b": "vb"}]],
            ["d[a=va]::b", ["vb"]],
            ["x::a::c", [1, 2, 3, 4]],
            ["x::c", [5, 6]],
            ["x::[c]", [{"c": 5}, {"c": 6}]],
            ["x::[c=5]", [{"c": 5}]],
            ["x::a[b]::c", [1, 2]],
            ["x::a[!b]::c", [3, 4]],
            ["x::a::l", [["l1"], ["l2"], ["l3"], ["l4"]]],
            [{"ref": "a[=$yes]", "vars": {"yes": "test"}}, ["test"]],
            [{"ref": "a[=$no]", "vars": {"no": None}}, []],
            ["[a]", [resource]],
            ["[=blah]", []],
            ["[blah]", []],
            ["[!blah]", [resource]],
            [".[!=blah]", [resource]],
            ["[!a]", []],
            ["::test", [resource]],
            ["d::*", set(["va", "vb"])],
            ["e::*::b2", ["v2"]],
            ["*", []],
            ["f", [{"a": 1, "b": 1}]],
            ["::*", [resource]],
            ["::*::.template::type", ["tosca.nodes.Root"]],
            # [{"q": "{{ foo }}"}, ["{{ foo }}"]]
            # XXX test nested ['.[k[d=3]=4]']
        ]:
            ref = Ref(exp)
            # print ('eval', ref.source, ref)
            if isinstance(expected, set):
                # for results where order isn't guaranteed in python2.7
                self.assertEqual(
                    set(ref.resolve(RefContext(resource))),
                    expected,
                    "expr was: " + ref.source,
                )
            else:
                self.assertEqual(
                    ref.resolve(RefContext(resource, trace=0)),
                    expected,
                    "expr was: " + ref.source,
                )

    def test_funcs(self):
        resource = self._getTestResource()
        test1 = {"ref": ".name", "vars": {"a": None}}
        test2 = {"ref": "$b", "vars": {"b": 1}}
        test3 = {
            "ref": {
                "if": {"not": "$a"},
                "then": {"q": "expected"},
                "else": {"q": "unexpected"},
            },
            "vars": {"a": None},
        }
        result1 = Ref(test1).resolveOne(RefContext(resource))
        self.assertEqual("test", result1)
        result2 = Ref(test2).resolveOne(RefContext(resource))
        self.assertEqual(1, result2)
        result3 = Ref(test3).resolveOne(RefContext(resource))
        self.assertEqual("expected", result3)
        result4 = Ref(test3).resolve(RefContext(resource))
        self.assertEqual(["expected"], result4)
        test5 = {"ref": {"or": ["$a", "b"]}, "vars": {"a": None}}
        result5 = Ref(test5).resolveOne(RefContext(resource))
        self.assertEqual(
            resource.attributes["b"], result5
        )  # this doesn't seem obvious!

    def test_forEach(self):
        resource = self._getTestResource()
        test1 = {"ref": ".", "foreach": {"value": {"content": {"ref": "b"}}}}
        expected0 = {"content": [1, 2, 3]}
        result0 = Ref(test1).resolveOne(RefContext(resource, trace=0))
        self.assertEqual(expected0, result0)
        # resolve has same result as resolveOne
        self.assertEqual([expected0], Ref(test1).resolve(RefContext(resource)))

        # add 'key' to make result a dict
        # test that template strings work
        # XXX fragile: key is base64 of __str__ of NodeInstance
        test1["foreach"]["key"] = "{{ item | ref | b64encode}}"
        result1 = Ref(test1).resolveOne(RefContext(resource))
        expected = {"Tm9kZUluc3RhbmNlKCd0ZXN0Jyk=": expected0}
        self.assertEqual(expected, result1, result1)
        result2 = Ref(test1).resolve(RefContext(resource))
        self.assertEqual([expected], result2)

    def test_serializeValues(self):
        resource = self._getTestResource()
        src = {"a": ["b", resource]}
        serialized = serializeValue(src)
        self.assertEqual(serialized, {"a": ["b", {"ref": "::test"}]})
        self.assertEqual(src, mapValue(serialized, resource))

    def test_jinjaTemplate(self):
        resource = NodeInstance("test", attributes=dict(a1="hello"))
        ctx = RefContext(resource, {"foo": "hello"})

        self.assertEqual(applyTemplate(" {{ foo }} ", ctx), "hello")
        # test jinja2 native types
        self.assertEqual(applyTemplate(" {{[foo]}} ", ctx), ["hello"])

        self.assertEqual(applyTemplate(' {{ "::test::a1" | ref }} ', ctx), u"hello")
        self.assertEqual(
            applyTemplate(' {{ lookup("unfurl", "::test::a1") }} ', ctx), u"hello"
        )
        # ansible query() always returns a list
        self.assertEqual(
            applyTemplate('{{  query("unfurl", "::test::a1") }}', ctx), [u"hello"]
        )

        os.environ[
            "TEST_ENV"
        ] = (
            "testEnv"
        )  # note: tox doesn't pass on environment variables so we need to set one now
        self.assertEqual(
            mapValue("{{ lookup('env', 'TEST_ENV') }}", resource), "testEnv"
        )

        # test that ref vars as can be used as template string vars
        exp = {"a": "{{ aVar }} world"}
        vars = {"aVar": "hello"}
        self.assertEqual(
            mapValue(exp, RefContext(resource, vars)), {"a": "hello world"}
        )

        vars = {"foo": {"bar": sensitive_str("sensitive")}}
        val = applyTemplate("{{ foo.bar }}", RefContext(resource, vars, trace=0))
        assert isinstance(val, sensitive_str), type(val)

    def test_templateFunc(self):
        query = {
            "eval": {"template": "{%if testVar %}{{success}}{%else%}failed{%endif%}"},
            "vars": {
                "testVar": True,
                "success": dict(eval={"if": "$true", "then": ".name"}),
            },
        }
        resource = self._getTestResource({"aTemplate": query})
        self.assertEqual(mapValue(query, resource), "test")
        self.assertEqual(resource.attributes["aTemplate"], "test")

        template = """\
#jinja2: variable_start_string: '<%', variable_end_string: '%>'
{% filter from_yaml %}
a_dict:
    key1: "{{ extras.quoted }}" # shouldn't be evaluated
    key2: "<% extras.quoted %>" # will be evaluated to "{{ quoted }}  "
    key3: "original"
    <<: <% key4 | mapValue | to_json %>  # will merge into a_dict
{% endfilter %}
        """
        vars = {
            "extras": dict(key3="hello", quoted={"q": "{{ quoted }}  "}),
            "key4": dict(extra1={"a": 1}, extra2=[1, 3], key3="overwritten"),
            "key2": {"q": "{{ quoted2 }}  "},
        }
        query2 = {"eval": {"template": template}}
        expected = {
            "a_dict": {
                "key1": "{{ extras.quoted }}",
                "key2": "{{ quoted }}  ",
                "key3": "original",
                "extra1": {"a": 1},
                "extra2": [1, 3],
            }
        }
        ctx = RefContext(resource, vars)
        result = applyTemplate(template, ctx)
        self.assertEqual(result, expected)
        result = mapValue(query2, ctx)
        self.assertEqual(result, expected)

    def test_innerReferences(self):
        resourceDef = {
            "a": dict(b={"ref": "a::c"}, c={"e": 1}, d=["2", {"ref": "a::d::0"}])
        }
        resource = NodeInstance("test", attributes=resourceDef)
        assert not not resource.attributes
        self.assertEqual(len(resource.attributes), 1)

        expectedA = {"c": {"e": 1}, "b": {"e": 1}, "d": ["2", "2"]}
        self.assertEqual(resource.attributes["a"]["b"], expectedA["b"])
        self.assertEqual(resource.attributes["a"], expectedA)
        self.assertEqual(Ref("a").resolve(RefContext(resource)), [expectedA])
        self.assertEqual(Ref("a").resolveOne(RefContext(resource)), expectedA)

        expected = ["2"]
        self.assertEqual(Ref("a::d::0").resolve(RefContext(resource)), expected)
        self.assertEqual(Ref("a::d::1").resolve(RefContext(resource)), expected)

        # print('test_references', resource.attributes,
        #   'AAAA', resource.attributes['a'],
        #   'BBB', resource.attributes['a']['b'],
        # )
        self.assertEqual(resource.attributes["a"], expectedA)
        self.assertEqual(resource.attributes["a"]["d"][0], "2")
        self.assertEqual(resource.attributes["a"]["d"][1], "2")
        self.assertEqual(resource.attributes["a"]["b"]["e"], 1)

        self.assertEqual(Ref("a::b::e").resolve(RefContext(resource)), [1])

        # test again to make sure it still resolves correctly
        self.assertEqual(Ref("a::d::0").resolve(RefContext(resource)), expected)
        self.assertEqual(Ref("a::d::1").resolve(RefContext(resource)), expected)

    def test_vars(self):
        # test dereferencing vars
        resource = self._getTestResource()
        query = {
            "eval": "$aDict",
            "vars": {"aDict": {"aRef": {"eval": "::test"}, "aTemplate": "{{ true }}"}},
        }
        result = Ref(query).resolveOne(RefContext(resource))
        self.assertEqual(result, {"aRef": resource, "aTemplate": True})

        query = {"eval": "$aRef", "vars": {"aRef": {"eval": "::test"}}}
        assert Ref.isRef(query["vars"]["aRef"])
        result = Ref(query).resolveOne(RefContext(resource))
        self.assertEqual(result, resource)

    def test_nodeTraversal1(self):
        root = NodeInstance(
            "r2", {"a": [dict(ref="::r1::a"), dict(ref="b")], "b": "r2"}  #'r1'  #'r2'
        )
        child = NodeInstance("r1", {"a": dict(ref="b"), "b": "r1"}, root)
        ctx = RefContext(root)
        x = [{"a": [{"c": 1}, {"c": 2}]}]
        assert x == ResultsList(x, ctx)
        self.assertEqual(Ref("b").resolve(RefContext(child)), ["r1"])
        self.assertEqual(Ref("a").resolve(RefContext(child)), ["r1"])
        self.assertEqual(Ref("a").resolve(RefContext(root)), [["r1", "r2"]])

    def test_nodeTraversal2(self):
        root = NodeInstance("root", {"a": [{"ref": "::child"}, {"b": 2}]})
        child = NodeInstance("child", {"b": 1}, root)
        self.assertEqual(Ref(".ancestors").resolve(RefContext(child)), [[child, root]])
        # self.assertEqual(Ref('a::b').resolve(RefContext(root)), [1])
        self.assertEqual(Ref("a").resolve(RefContext(child)), [[child, {"b": 2}]])
        # a resolves to [child, dict] so a::b resolves to [child[b], [b]2]
        self.assertEqual(Ref("a::b").resolve(RefContext(child)), [1, 2])

    def test_lookup(self):
        resource = self._getTestResource()
        os.environ[
            "TEST_ENV"
        ] = (
            "testEnv"
        )  # note: tox doesn't pass on environment variables so we need to set one now
        query = {"eval": {"lookup": {"env": "TEST_ENV"}}}
        self.assertEqual(mapValue(query, resource), "testEnv")

    def test_tempfile(self):
        resource = self._getTestResource()
        value = {"a": 1}
        template = dict(
            eval={"template": "{{ valuesfile }}"},
            vars={"valuesfile": {"eval": {"tempfile": value}}},
        )
        result = mapValue(template, resource)
        with open(result) as tp:
            self.assertEqual(tp.read(), json.dumps(value, indent=2))

    def test_binaryvault(self):
        import six
        from unfurl.support import AttributeManager
        from unfurl.yamlloader import makeYAML, sensitive_bytes, makeVaultLib

        # load a binary file then write it out as a temporary vault file
        fixture = os.path.join(
            os.path.dirname(__file__), "fixtures/helmrepo/mysql-1.6.4.tgz"
        )
        src = (
            """
          eval:
            tempfile:
                eval:
                  file: %s
                select: contents
            encoding: vault
        """
            % fixture
        )
        vault = makeVaultLib("password")
        yaml = makeYAML(vault)
        expr = yaml.load(six.StringIO(src))
        resource = self._getTestResource()
        resource.attributeManager = AttributeManager(yaml)
        resource._templar._loader.set_vault_secrets(vault.secrets)

        filePath = mapValue(expr, resource)
        with open(filePath, "rb") as vf:
            vaultContents = vf.read()
            assert vaultContents.startswith(b"$ANSIBLE_VAULT;")

        # decrypt the vault file, make sure it's a sensitive_bytes string that matches the original contents
        src = """
        eval:
          file:
            eval: $tempfile
          encoding: binary
        select: contents
        """
        expr = yaml.load(six.StringIO(src))
        contents = mapValue(expr, RefContext(resource, vars=dict(tempfile=filePath)))
        assert isinstance(contents, sensitive_bytes), type(contents)
        with open(fixture, "rb") as tp:
            self.assertEqual(tp.read(), contents)



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
from __future__ import absolute_import
import os.path

import six
from six.moves.configparser import ConfigParser

from supervisor import xmlrpc

try:
    from xmlrpc.client import ServerProxy, Fault
except ImportError:
    from xmlrpclib import Server as ServerProxy
    from xmlrpclib import Fault

from ..configurator import Configurator
from ..support import abspath

# support unix domain socket connections
# (we want to connect the same way as supervisorctl does for security and to automatically support multiple instances)
def getServerProxy(serverurl=None, username=None, password=None):
    # copied from https://github.com/Supervisor/supervisor/blob/b52f49cff287c4d821c2c54d7d1afcd397b699e5/supervisor/options.py#L1718
    return ServerProxy(
        # dumbass ServerProxy won't allow us to pass in a non-HTTP url,
        # so we fake the url we pass into it and always use the transport's
        # 'serverurl' to figure out what to attach to
        "http://127.0.0.1",
        transport=xmlrpc.SupervisorTransport(username, password, serverurl),
    )


def _reloadConfig(server, name):
    result = server.supervisor.reloadConfig()
    return any((name in changed) for changed in result[0])


class SupervisorConfigurator(Configurator):
    def run(self, task):
        host = task.vars["HOST"]

        # if homeDir is a relative path it will be relative to the baseDir of the host instance
        # which might be different from the current directory if host is an external instance
        confDir = abspath(host.context, host["homeDir"])
        conf = host["conf"]

        name = task.vars["SELF"]["name"]
        confPath = os.path.join(confDir, "programs", name + ".conf")
        if six.PY3:
            parser = ConfigParser(inline_comment_prefixes=(";", "#"), strict=False)
            parser.read_string(conf)
        else:
            parser = ConfigParser()
            parser.readfp(six.StringIO(conf))
        serverConfig = dict(parser.items("supervisorctl", vars=dict(here=confDir)))
        serverConfig.pop("here", None)
        server = getServerProxy(**serverConfig)

        error = None
        op = task.configSpec.operation
        modified = False
        try:
            if op == "start":
                server.supervisor.startProcess(name)
                modified = True
            elif op == "stop":
                server.supervisor.stopProcess(name)
                modified = True
            elif op == "delete":
                if os.path.exists(confPath):
                    os.remove(confPath)
                modified = _reloadConfig(server, name)
            elif op == "configure":
                program = task.vars["SELF"]["program"]
                programDir = os.path.dirname(confPath)
                task.logger.debug("writing %s", confPath)
                if not os.path.isdir(programDir):
                    os.makedirs(programDir)
                with open(confPath, "w") as conff:
                    conf = "[program:%s]\n" % name
                    conf += "\n".join("%s= %s" % (k, v) for k, v in program.items())
                    if "environment" not in program:
                        conf += "\nenvironment= "
                        conf += ",".join(
                            '%s="%s"' % (k, v.replace("%", "%%"))
                            for (k, v) in task.getEnvironment(True).items()
                        )
                    conff.write(conf)
                modified = _reloadConfig(server, name)
                server.supervisor.addProcessGroup(name)
        except Fault as err:
            if (
                not (op == "start" and err.faultCode == 60)  # ok, 60 == ALREADY_STARTED
                and not (op == "stop" and err.faultCode == 70)  # ok, 70 == NOT_RUNNING
                and not (  # ok, 90 == ALREADY_ADDED
                    op == "configure" and err.faultCode == 90
                )
            ):
                error = "supervisor error: " + str(err)

        yield task.done(success=not error, modified=modified, result=error)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging

from toscaparser.entity_template import EntityTemplate
from toscaparser.properties import Property

log = logging.getLogger('tosca')


class RelationshipTemplate(EntityTemplate):
    '''Relationship template.'''

    SECTIONS = (DERIVED_FROM, PROPERTIES, REQUIREMENTS,
                INTERFACES, TYPE, DEFAULT_FOR) = \
               ('derived_from', 'properties', 'requirements', 'interfaces',
                 'type', 'default_for')
    ANY = 'ANY'

    def __init__(self, relationship_template, name, custom_def=None,
                 target=None, source=None):
        super(RelationshipTemplate, self).__init__(name,
                                                   relationship_template,
                                                   'relationship_type',
                                                   custom_def)
        self.name = name
        self.target = target
        self.source = source
        self.capability = None
        self.default_for = self.entity_tpl.get(self.DEFAULT_FOR)

    def get_matching_capabilities(self, targetNodeTemplate, capability_name=None):
        capabilitiesDict = targetNodeTemplate.get_capabilities()
        # if capability_name is set, make sure the target node has a capability
        # that matching it as a name or or as a type
        if capability_name:
            capability = capabilitiesDict.get(capability_name)
            if capability:
                # just test the capability that matches the symbolic name
                capabilities = [capability]
            else:
                # name doesn't match a symbolic name, see if its a valid type name
                capabilities = [cap for cap in capabilitiesDict.values() if cap.is_derived_from(capability_name)]
        else:
            capabilities = list(capabilitiesDict.values())

        # if valid_target_types is set, make sure the matching capabilities are compatible
        capabilityTypes = self.type_definition.valid_target_types
        if capabilityTypes:
            capabilities = [cap for cap in capabilities
                              if any(cap.is_derived_from(capType) for capType in capabilityTypes)]
        if not capability_name and len(capabilities) > 1:
            # if no capability was specified and there are more than one to choose from, choose the most generic
            featureCap = capabilitiesDict.get("feature")
            if featureCap:
                return [featureCap]
        return capabilities



# Copyright (c) 2020 Adam Souzis
# SPDX-License-Identifier: MIT
"""
Classes for managing the local environment.

Repositories can optionally be organized into projects that have a local configuration.

By convention, the "home" project defines a localhost instance and adds it to its context.
"""
import os
import os.path
import datetime

import six
from .repo import Repo, normalizeGitUrl, splitGitUrl, RepoView
from .util import UnfurlError
from .merge import mergeDicts
from .yamlloader import YamlConfig, makeVaultLib
from . import DefaultNames, getHomeConfigPath
from six.moves.urllib.parse import urlparse

_basepath = os.path.abspath(os.path.dirname(__file__))


class Project(object):
    """
    A Unfurl project is a folder that contains at least a local configuration file (unfurl.yaml),
    one or more ensemble.yaml files which maybe optionally organized into one or more git repositories.
    """

    def __init__(self, path, homeProject=None):
        assert isinstance(path, six.string_types), path
        parentConfig = homeProject and homeProject.localConfig or None
        self.projectRoot = os.path.abspath(os.path.dirname(path))
        if os.path.exists(path):
            self.localConfig = LocalConfig(path, parentConfig)
        else:
            self.localConfig = LocalConfig(parentConfig=parentConfig)
        self._setRepos()
        if self.projectRepo and homeProject:
            homeProject.localConfig.registerProject(self)

    def _setRepos(self):
        # abspath => RepoView:
        self.workingDirs = Repo.findGitWorkingDirs(self.projectRoot)
        # the project repo if it exists manages the project config (unfurl.yaml)
        projectRoot = self.projectRoot
        if projectRoot in self.workingDirs:
            self.projectRepo = self.workingDirs[projectRoot].repo
        else:
            # project maybe part of a containing repo (if created with --existing option)
            repo = Repo.findContainingRepo(self.projectRoot)
            # make sure projectroot isn't excluded from the containing repo
            if repo and not repo.isPathExcluded(self.projectRoot):
                self.projectRepo = repo
                self.workingDirs[repo.workingDir] = repo.asRepoView()
            else:
                self.projectRepo = None

        if self.projectRepo:
            # Repo.findGitWorkingDirs() doesn't look inside git working dirs
            # so look for repos in dirs that might be excluded from git
            for dir in self.projectRepo.findExcludedDirs(self.projectRoot):
                if projectRoot in dir and os.path.isdir(dir):
                    Repo.updateGitWorkingDirs(self.workingDirs, dir, os.listdir(dir))

        # add referenced local repositories outside of the project
        for path in self.localConfig.localRepositories:
            if os.path.isdir(path):
                # XXX assumes its a git repo, should compare and validate lock metadata
                Repo.updateGitWorkingDirs(self.workingDirs, path, os.listdir(path))

    @staticmethod
    def normalizePath(path):
        path = os.path.abspath(path)
        if not os.path.exists(path):
            isdir = not path.endswith(".yml") and not path.endswith(".yaml")
        else:
            isdir = os.path.isdir(path)

        if isdir:
            return os.path.join(path, DefaultNames.LocalConfig)
        else:
            return path

    @staticmethod
    def findPath(testPath):
        """
        Walk parents looking for unfurl.yaml
        """
        current = os.path.abspath(testPath)
        while current and current != os.sep:
            test = os.path.join(current, DefaultNames.LocalConfig)
            if os.path.exists(test):
                return test
            current = os.path.dirname(current)
        return None

    @property
    def venv(self):
        venv = os.path.join(self.projectRoot, ".venv")
        if os.path.isdir(venv):
            return venv
        return None

    def getAsdfPaths(self, asdfDataDir, toolVersions={}):
        paths = []
        toolVersionsFilename = (
            os.getenv("ASDF_DEFAULT_TOOL_VERSIONS_FILENAME") or ".tool-versions"
        )
        versionConf = os.path.join(self.projectRoot, toolVersionsFilename)
        if os.path.exists(versionConf):
            with open(versionConf) as conf:
                for line in conf.readlines():
                    line = line.strip()
                    if line and line[0] != "#":
                        tokens = line.split()
                        plugin = tokens.pop(0)
                        versions = toolVersions.setdefault(plugin, set())
                        for version in tokens:
                            if version != "system":
                                path = os.path.join(
                                    asdfDataDir, "installs", plugin, version, "bin"
                                )
                                versions.add(version)
                                if os.path.isdir(path):
                                    paths.append(path)
        return paths

    def getRepos(self):
        repos = [r.repo for r in self.workingDirs.values()]
        if self.projectRepo and self.projectRepo not in repos:
            repos.append(self.projectRepo)
        return repos

    def getDefaultEnsemblePath(self):
        fullPath = self.localConfig.getDefaultManifestPath()
        if fullPath:
            return fullPath
        return os.path.join(
            self.projectRoot, DefaultNames.EnsembleDirectory, DefaultNames.Ensemble
        )

    def findDefaultInstanceManifest(self):
        fullPath = self.localConfig.getDefaultManifestPath()
        if fullPath:
            if not os.path.exists(fullPath):
                raise UnfurlError(
                    "The default ensemble found in %s does not exist: %s"
                    % (self.localConfig.config.path, os.path.abspath(fullPath))
                )
            return fullPath
        else:
            # no manifest specified in the project config so check the default locations
            fullPath = os.path.join(
                self.projectRoot, DefaultNames.EnsembleDirectory, DefaultNames.Ensemble
            )
            if os.path.exists(fullPath):
                return fullPath
            fullPath2 = os.path.join(self.projectRoot, DefaultNames.Ensemble)
            if os.path.exists(fullPath2):
                return fullPath2
            raise UnfurlError(
                'The can not find an ensemble in a default location: "%s" or "%s"'
                % (fullPath, fullPath2)
            )

    def getRelativePath(self, path):
        return os.path.relpath(os.path.abspath(path), self.projectRoot)

    def isPathInProject(self, path):
        return not self.getRelativePath(path).startswith("..")

    def _createPathForGitRepo(self, gitUrl):
        name = Repo.getPathForGitRepo(gitUrl)
        return self.getUniquePath(name)

    def getUniquePath(self, name):
        basename = name
        counter = 1
        while os.path.exists(os.path.join(self.projectRoot, name)):
            name = basename + str(counter)
            counter += 1
        return os.path.join(self.projectRoot, name)

    def findGitRepo(self, repoURL, revision=None):
        candidate = None
        for dir, repository in self.workingDirs.items():
            repo = repository.repo
            if repoURL.startswith("git-local://"):
                initialCommit = urlparse(repoURL).netloc.partition(":")[0]
                match = initialCommit == repo.getInitialRevision()
            else:
                match = normalizeGitUrl(repoURL) == normalizeGitUrl(repo.url)
            if match:
                if revision:
                    if repo.revision == repo.resolveRevSpec(revision):
                        return repo
                    candidate = repo
                else:
                    return repo
        return candidate

    def findPathInRepos(self, path, importLoader=None):
        """If the given path is part of the working directory of a git repository
        return that repository and a path relative to it"""
        # importloader is unused until pinned revisions are supported
        candidate = None
        for dir in sorted(self.workingDirs.keys()):
            repo = self.workingDirs[dir].repo
            filePath = repo.findRepoPath(path)
            if filePath is not None:
                return repo, filePath, repo.revision, False
            #  XXX support bare repo and particular revisions
            #  need to make sure path isn't ignored in repo or compare candidates
            # filePath, revision, bare = repo.findPath(path, importLoader)
            # if filePath is not None:
            #     if not bare:
            #         return repo, filePath, revision, bare
            #     else:  # if it's bare see if we can find a better candidate
            #         candidate = (repo, filePath, revision, bare)
        return candidate or None, None, None, None

    def createWorkingDir(self, gitUrl, ref=None):
        localRepoPath = self._createPathForGitRepo(gitUrl)
        repo = Repo.createWorkingDir(gitUrl, localRepoPath, ref)
        # add to workingDirs
        self.workingDirs[os.path.abspath(localRepoPath)] = repo.asRepoView()
        return repo

    def findGitRepoFromRepository(self, repoSpec):
        repoUrl = repoSpec.url
        return self.findGitRepo(splitGitUrl(repoUrl)[0])

    def findOrClone(self, repo):
        gitUrl = repo.url
        existingRepo = self.findGitRepo(gitUrl)
        if existingRepo:
            return existingRepo

        # if not found:
        localRepoPath = os.path.abspath(
            self._createPathForGitRepo(repo.workingDir or gitUrl)
        )
        newRepo = repo.clone(localRepoPath)
        # use gitUrl to preserve original origin
        self.workingDirs[localRepoPath] = RepoView(dict(name="", url=gitUrl), newRepo)
        return newRepo


class LocalConfig(object):
    """
    Represents the local configuration file, which provides the environment that manifests run in, including:
      instances imported from other ensembles, inputs, environment variables, secrets and local configuration.

    It consists of:
    * a list of ensemble manifests with their local configuration
    * the default local and secret instances"""

    # don't merge the value of the keys of these dicts:
    replaceKeys = [
        "inputs",
        "attributes",
        "schemas",
        "connections",
        "manifest",
        "environment",
    ]

    def __init__(self, path=None, parentConfig=None, validate=True):
        defaultConfig = {"apiVersion": "unfurl/v1alpha1", "kind": "Project"}
        self.config = YamlConfig(
            defaultConfig, path, validate, os.path.join(_basepath, "unfurl-schema.json")
        )
        self.manifests = self.config.expanded.get("manifests") or []
        self.projects = self.config.expanded.get("projects") or {}
        contexts = self.config.expanded.get("contexts", {})
        if parentConfig:
            parentContexts = parentConfig.config.expanded.get("contexts", {})
            contexts = mergeDicts(
                parentContexts, contexts, replaceKeys=self.replaceKeys
            )
        self.contexts = contexts
        self.parentConfig = parentConfig
        self.localRepositories = self.config.expanded.get("localRepositories") or {}

    def getContext(self, manifestPath, context):
        localContext = self.contexts.get("defaults") or {}
        contextName = "defaults"
        for spec in self.manifests:
            if manifestPath == self.adjustPath(spec["file"]):
                # use the context associated with the given manifest
                contextName = spec.get("context", contextName)
                break

        if contextName != "defaults" and self.contexts.get(contextName):
            localContext = mergeDicts(
                localContext, self.contexts[contextName], replaceKeys=self.replaceKeys
            )

        return mergeDicts(context, localContext, replaceKeys=self.replaceKeys)

    def adjustPath(self, path):
        """
        Makes sure relative paths are relative to the location of this local config
        """
        return os.path.join(self.config.getBaseDir(), path)

    def getDefaultManifestPath(self):
        if len(self.manifests) == 1:
            return self.adjustPath(self.manifests[0]["file"])
        else:
            for spec in self.manifests:
                if spec.get("default"):
                    return self.adjustPath(spec["file"])
        return None

    def createLocalInstance(self, localName, attributes):
        # local or secret
        from .runtime import NodeInstance

        if "default" in attributes:
            if not "default" in attributes.get(".interfaces", {}):
                attributes.setdefault(".interfaces", {})[
                    "default"
                ] = "unfurl.support.DelegateAttributes"
        if "inheritFrom" in attributes:
            if not "inherit" in attributes.get(".interfaces", {}):
                attributes.setdefault(".interfaces", {})[
                    "inherit"
                ] = "unfurl.support.DelegateAttributes"
        instance = NodeInstance(localName, attributes)
        instance._baseDir = self.config.getBaseDir()
        return instance

    def registerProject(self, project):
        # update, if necessary, localRepositories and projects
        key, local = self.config.searchIncludes(key="localRepositories")
        if not key and "localRepositories" not in self.config.config:
            # localRepositories doesn't exist, see if we are including a file inside "local"
            pathPrefix = os.path.join(self.config.getBaseDir(), "local")
            key, local = self.config.searchIncludes(pathPrefix=pathPrefix)
        if not key:
            local = self.config.config

        repo = project.projectRepo
        localRepositories = local.setdefault("localRepositories", {})
        lock = repo.asRepoView().lock()
        if localRepositories.get(repo.workingDir) != lock:
            localRepositories[repo.workingDir] = lock
        else:
            return False  # no change

        if not repo.isLocalOnly():
            for name, val in self.projects.items():
                if normalizeGitUrl(val["url"]) == normalizeGitUrl(repo.url):
                    break
            else:
                # if project isn't already in projects, use generated name
                name = "_" + os.path.basename(project.projectRoot)
                counter = 0
                while name in self.projects:
                    counter += 1
                    name += "-%s" % counter

            externalProject = dict(
                url=repo.url,
                initial=repo.getInitialRevision(),
            )
            file = os.path.relpath(project.projectRoot, repo.workingDir)
            if file and file != ".":
                externalProject["file"] = file

            self.projects[name] = externalProject
            self.config.config.setdefault("projects", {})[name] = externalProject

        if key:
            self.config.saveInclude(key)
        self.config.save()
        return True


class LocalEnv(object):
    """
    This class represents the local environment that an ensemble runs in, including
    the local project it is part of and the home project.
    """

    homeProject = None

    def __init__(self, manifestPath=None, homePath=None, parent=None, project=None):
        """
        If manifestPath is None find the first unfurl.yaml or ensemble.yaml
        starting from the current directory.

        If homepath is set it overrides UNFURL_HOME
        (and an empty string disable the home path).
        Otherwise the home path will be set to UNFURL_HOME or the default home location.
        """
        import logging

        logger = logging.getLogger("unfurl")
        self.logger = logger

        if parent:
            self._projects = parent._projects
            self._manifests = parent._manifests
            self.homeConfigPath = parent.homeConfigPath
        else:
            self._projects = {}
            if project:
                self._projects[project.localConfig.config.path] = project
            self._manifests = {}
            self.homeConfigPath = getHomeConfigPath(homePath)

        if self.homeConfigPath:
            if not os.path.exists(self.homeConfigPath):
                logger.warning(
                    'UNFURL_HOME is set but does not exist: "%s"', self.homeConfigPath
                )
            else:
                self.homeProject = self.getProject(self.homeConfigPath, None)
                if not self.homeProject:
                    logger.warning(
                        'Could not load home project at: "%s"', self.homeConfigPath
                    )

        self.manifestPath = None
        if manifestPath:
            # if manifestPath does not exist check project config
            manifestPath = os.path.abspath(manifestPath)
            if not os.path.exists(manifestPath):
                # XXX check if the manifest is named in the project config
                # pathORproject = self.findProject(os.path.dirname(manifestPath))
                # if pathORproject:
                #    self.manifestPath = pathORproject.getInstance(manifestPath)
                # else:
                raise UnfurlError(
                    "Ensemble manifest does not exist: '%s'" % manifestPath
                )
            else:
                pathORproject = self.findManifestPath(manifestPath)
        else:
            # not specified: search current directory and parents for either a manifest or a project
            pathORproject = self.searchForManifestOrProject(".")

        if isinstance(pathORproject, Project):
            self.project = pathORproject
            if not self.manifestPath:
                self.manifestPath = pathORproject.findDefaultInstanceManifest()
        else:
            self.manifestPath = pathORproject
            if project:
                self.project = project
            else:
                self.project = self.findProject(os.path.dirname(pathORproject))

        self.toolVersions = {}
        self.instanceRepo = self._getInstanceRepo()
        self.config = (
            self.project
            and self.project.localConfig
            or self.homeProject
            and self.homeProject.localConfig
            or LocalConfig()
        )

    def getVaultPassword(self, vaultId="default"):
        secret = os.getenv("UNFURL_VAULT_%s_PASSWORD" % vaultId.upper())
        if not secret:
            context = self.getContext()
            secret = (
                context.get("secrets", {})
                .get("attributes", {})
                .get("vault_%s_password" % vaultId)
            )
        return secret

    def getManifest(self, path=None):
        from .yamlmanifest import YamlManifest

        if path and path != self.manifestPath:
            # share projects and ensembles
            localEnv = LocalEnv(path, parent=self)
            return localEnv.getManifest()
        else:
            manifest = self._manifests.get(self.manifestPath)
            if not manifest:
                vaultId = "default"
                vault = makeVaultLib(self.getVaultPassword(vaultId), vaultId)
                if vault:
                    self.logger.info(
                        "Vault password found, configuring vault id: %s", vaultId
                    )
                manifest = YamlManifest(localEnv=self, vault=vault)
                self._manifests[self.manifestPath] = manifest
            return manifest

    def getProject(self, path, homeProject):
        path = Project.normalizePath(path)
        project = self._projects.get(path)
        if not project:
            project = Project(path, homeProject)
            self._projects[path] = project
        return project

    def getExternalManifest(self, location):
        assert "project" in location
        project = None
        if self.project:
            project = self.project.localConfig.projects.get(location["project"])
        if not project and self.homeProject:
            project = self.homeProject.localConfig.projects.get(location["project"])
        if not project:
            return None
        repo = self.findGitRepo(project["url"])
        if not repo:
            return None

        projectRoot = os.path.join(repo.workingDir, project.get("file") or "")
        localEnv = LocalEnv(
            os.path.join(projectRoot, location.get("file") or ""), parent=self
        )
        return localEnv.getManifest()

    # manifestPath specified
    #  doesn't exist: error
    #  is a directory: either instance repo or a project
    def findManifestPath(self, manifestPath):
        if not os.path.exists(manifestPath):
            raise UnfurlError(
                "Manifest file does not exist: '%s'" % os.path.abspath(manifestPath)
            )

        if os.path.isdir(manifestPath):
            test = os.path.join(manifestPath, DefaultNames.Ensemble)
            if os.path.exists(test):
                return test
            else:
                test = os.path.join(manifestPath, DefaultNames.LocalConfig)
                if os.path.exists(test):
                    return self.getProject(test, self.homeProject)
                else:
                    message = (
                        "Can't find an Unfurl ensemble or project in folder '%s'"
                        % manifestPath
                    )
                    raise UnfurlError(message)
        else:
            return manifestPath

    def _getInstanceRepo(self):
        instanceDir = os.path.dirname(self.manifestPath)
        if self.project and instanceDir in self.project.workingDirs:
            return self.project.workingDirs[instanceDir].repo
        else:
            return Repo.findContainingRepo(instanceDir)

    def getRepos(self):
        if self.project:
            repos = self.project.getRepos()
        else:
            repos = []
        if self.instanceRepo and self.instanceRepo not in repos:
            return repos + [self.instanceRepo]
        else:
            return repos

    def searchForManifestOrProject(self, dir):
        current = os.path.abspath(dir)
        while current and current != os.sep:
            test = os.path.join(current, DefaultNames.Ensemble)
            if os.path.exists(test):
                return test

            test = os.path.join(current, DefaultNames.LocalConfig)
            if os.path.exists(test):
                return self.getProject(test, self.homeProject)

            current = os.path.dirname(current)

        message = "Can't find an Unfurl ensemble or project in the current directory (or any of the parent directories)"
        raise UnfurlError(message)

    def findProject(self, testPath):
        """
        Walk parents looking for unfurl.yaml
        """
        path = Project.findPath(testPath)
        if path is not None:
            return self.getProject(path, self.homeProject)
        return None

    def getContext(self, context=None):
        """
        Return a new context that merges the given context with the local context.
        """
        return self.config.getContext(self.manifestPath, context or {})

    def getRuntime(self):
        context = self.getContext()
        runtime = context.get("runtime")
        if runtime:
            return runtime
        if self.project and self.project.venv:
            return "venv:" + self.project.venv
        if self.homeProject and self.homeProject.venv:
            return "venv:" + self.homeProject.venv
        return None

    def getLocalInstance(self, name, context):
        assert name in ["locals", "secrets", "local", "secret"]
        local = context.get(name, {})
        return (
            self.config.createLocalInstance(
                name.rstrip("s"), local.get("attributes", {})
            ),
            local,
        )

    def findGitRepo(self, repoURL, revision=None):
        repo = None
        if self.project:
            repo = self.project.findGitRepo(repoURL, revision)
        if not repo:
            if self.homeProject:
                return self.homeProject.findGitRepo(repoURL, revision)
        return repo

    def findOrCreateWorkingDir(self, repoURL, revision=None, basepath=None):
        repo = self.findGitRepo(repoURL, revision)
        # git-local repos must already exist
        if not repo and not repoURL.startswith("git-local://"):
            if self.project and (
                basepath is None or self.project.isPathInProject(basepath)
            ):
                project = self.project
            else:
                project = self.homeProject
            if project:
                repo = project.createWorkingDir(repoURL, revision)
        if not repo:
            return None, None, None

        if revision:
            # bare if HEAD isn't at the requested revision
            bare = repo.revision != repo.resolveRevSpec(revision)
            return repo, repo.revision, bare
        else:
            return repo, repo.revision, False

    def findPathInRepos(self, path, importLoader=None):
        """If the given path is part of the working directory of a git repository
        return that repository and a path relative to it"""
        # importloader is unused until pinned revisions are supported
        if self.instanceRepo:
            repo = self.instanceRepo
            filePath = repo.findRepoPath(path)
            if filePath is not None:
                return repo, filePath, repo.revision, False

        candidate = None
        repo = None
        if self.project:
            repo, filePath, revision, bare = self.project.findPathInRepos(
                path, importLoader
            )
            if repo:
                if not bare:
                    return repo, filePath, revision, bare
                else:
                    candidate = (repo, filePath, revision, bare)

        if self.homeProject:
            repo, filePath, revision, bare = self.homeProject.findPathInRepos(
                path, importLoader
            )
        if repo:
            if bare and candidate:
                return candidate
            else:
                return repo, filePath, revision, bare
        return None, None, None, None

    def mapValue(self, val):
        """
        Evaluate using project home as a base dir.
        """
        from .runtime import NodeInstance
        from .eval import mapValue

        instance = NodeInstance()
        instance._baseDir = self.config.config.getBaseDir()
        return mapValue(val, instance)

    def getPaths(self):
        paths = []
        asdfDataDir = os.getenv("ASDF_DATA_DIR")
        if not asdfDataDir:
            # check if an ensemble previously installed asdf via git
            repo = self.findGitRepo("https://github.com/asdf-vm/asdf.git/")
            if repo:
                asdfDataDir = repo.workingDir
            else:
                homeAsdf = os.path.expanduser("~/.asdf")
                if os.path.isdir(homeAsdf):
                    asdfDataDir = homeAsdf
        if asdfDataDir:
            # project has higher priority over home project
            if self.project:
                paths = self.project.getAsdfPaths(asdfDataDir, self.toolVersions)
            if self.homeProject:
                paths += self.homeProject.getAsdfPaths(asdfDataDir, self.toolVersions)
        return paths
